# SPDX-FileCopyrightText: 2024-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
from tempfile import NamedTemporaryFile
from xml.etree import ElementTree

import pytest
from conftest import (
    create_project,
)

from idf_build_apps import App
from idf_build_apps.args import (
    BuildArguments,
    DependencyDrivenBuildArguments,
    FindArguments,
    FindBuildArguments,
)
from idf_build_apps.constants import IDF_BUILD_APPS_TOML_FN, PREVIEW_TARGETS, SUPPORTED_TARGETS
from idf_build_apps.main import main


def test_init_attr_deprecated_by():
    args = DependencyDrivenBuildArguments(
        ignore_app_dependencies_filepatterns=['bar'],
        modified_files=['barfile'],
    )
    assert args.deactivate_dependency_driven_build_by_filepatterns == ['bar']

    args = DependencyDrivenBuildArguments(
        deactivate_dependency_driven_build_by_components=['foo'],
        modified_components=['foocomp'],
    )
    assert args.deactivate_dependency_driven_build_by_components == ['foo']


@pytest.mark.parametrize(
    'kwargs, expected',
    [
        ({'config': 'foo'}, ['foo']),
        ({'config_rules_str': 'bar'}, ['bar']),
        ({'config_rules': ['baz']}, ['baz']),
    ],
)
def test_init_attr_override(kwargs, expected):
    args = FindBuildArguments(
        **kwargs,
    )
    assert args.config_rules == expected


def test_apply_config_with_deprecated_names():
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
    "foo"
]
""")

    args = FindBuildArguments()
    assert args.config_rules == ['foo']


def test_apply_config_in_parent_dir(tmp_path):
    test_under = tmp_path / 'test_under'
    test_under.mkdir()
    os.chdir(test_under)

    with open(tmp_path / IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write('target = "esp32"')

    assert FindArguments().target == 'esp32'


def test_apply_config_over_pyproject_toml(tmp_path):
    test_under = tmp_path / 'test_under'
    test_under.mkdir()
    os.chdir(test_under)

    with open(test_under / 'pyproject.toml', 'w') as fw:
        fw.write("""[tool.idf-build-apps]
target = "esp32s2"
""")

    assert FindArguments().target == 'esp32s2'

    with open(tmp_path / IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write('target = "esp32"')

    assert FindArguments().target == 'esp32'


def test_empty_argument():
    args = FindArguments()
    assert args.config_rules is None


def test_build_args_expansion():
    args = BuildArguments(
        parallel_index=2, collect_app_info='@p.txt', junitxml='x_@p.txt', collect_size_info='@p_@p.txt'
    )
    assert args.collect_app_info == '2.txt'
    assert args.junitxml == 'x_2.txt'

    args.parallel_index = 3
    assert args.collect_app_info == '3.txt'
    assert args.junitxml == 'x_3.txt'
    assert args.collect_size_info == '3_3.txt'


def test_func_overwrite_config():
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
        "foo"
    ]
    modified_components = [
        'comp1',
    ]
    no_color = true
    """)

    args = FindArguments(
        config=['bar'],
    )

    assert args.config_rules == ['bar']
    assert args.modified_components == ['comp1']
    assert args.modified_files is None


def test_func_overwrite_toml_overwrite_pyproject_toml():
    with open('pyproject.toml', 'w') as fw:
        fw.write("""[tool.idf-build-apps]
config = [
    "foo"
]
modified_components = [
    'comp1',
]
ignore_app_dependencies_components = [
    'baz'
]
verbose=3
    """)

    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
    "bar"
]
modified_files = [
    'file1',
]
        """)

    args = FindArguments(
        modified_components=['comp2'],
    )
    assert args.config_rules == ['bar']
    assert args.modified_components == ['comp2']
    assert args.modified_files == ['file1']
    assert args.verbose == 3
    assert args.deactivate_dependency_driven_build_by_components == ['baz']


def test_build_targets_cli(tmp_path, monkeypatch):
    create_project('foo', tmp_path)
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""paths = ["foo"]
build_dir = "build_@t"
junitxml = "test.xml"
dry_run = true
""")

    def get_enabled_targets(args):
        with monkeypatch.context() as m:
            m.setattr('sys.argv', args)
            main()
        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        testcases = test_suite.findall('testcase')
        # get targets
        return [c.attrib['name'].replace('foo/build_', '') for c in testcases]

    # default build SUPPORTED_TARGETS
    targets = get_enabled_targets(['idf-build-apps', 'build'])
    assert len(targets) == len(SUPPORTED_TARGETS)
    assert set(targets) == set(SUPPORTED_TARGETS)
    # build with --target
    targets = get_enabled_targets(['idf-build-apps', 'build', '--target', 'esp32'])
    assert len(targets) == 1
    assert targets[0] == 'esp32'
    # build with --default-build-targets
    targets = get_enabled_targets(['idf-build-apps', 'build', '--default-build-targets', 'esp32', 'esp32s2'])
    assert len(targets) == 2
    assert set(targets) == {'esp32', 'esp32s2'}
    # build with --enable-preview-targets
    targets = get_enabled_targets(['idf-build-apps', 'build', '--enable-preview-targets'])
    assert len(targets) == len(SUPPORTED_TARGETS) + len(PREVIEW_TARGETS)
    assert set(targets) == set(SUPPORTED_TARGETS) | set(PREVIEW_TARGETS)
    # build with --disable-targets
    assert 'esp32' in SUPPORTED_TARGETS and 'esp32s2' in SUPPORTED_TARGETS
    targets = get_enabled_targets(['idf-build-apps', 'build', '--disable-targets', 'esp32', 'esp32s2'])
    assert set(targets) == set(SUPPORTED_TARGETS) - {'esp32', 'esp32s2'}
    # build with --enable-preview-targets and --disable-targets
    targets = get_enabled_targets(
        ['idf-build-apps', 'build', '--enable-preview-targets', '--disable-targets', PREVIEW_TARGETS[0]]
    )
    assert len(targets) == len(SUPPORTED_TARGETS) + len(PREVIEW_TARGETS) - 1
    assert set(targets) == set(SUPPORTED_TARGETS) | set(PREVIEW_TARGETS) - {PREVIEW_TARGETS[0]}


class TestIgnoreWarningFile:
    def test_deprecated_cli(self, monkeypatch, capsys):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-file', 'foo.txt'])
            main()

        assert len(App.IGNORE_WARNS_REGEXES) == 1
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'

        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-file', 'foo.txt', 'bar.txt'])
            with pytest.raises(SystemExit):
                main()

        assert 'unrecognized arguments: bar.txt' in capsys.readouterr().err

    def test_new_cli(self, monkeypatch):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-files', 'foo.txt', 'bar.txt'])
            main()

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_func_with_str(self):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        BuildArguments(
            ignore_warning_files=['foo.txt', 'bar.txt'],
        )

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_func_with_fs(self):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        BuildArguments(
            ignore_warning_file=[open('foo.txt'), open('bar.txt')],
        )

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_ignore_extra_fields(self):
        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""dry_run = true""")

        args = FindArguments()
        assert not hasattr(args, 'dry_run')

    def test_config_file(self, tmp_path, monkeypatch):
        create_project('foo', tmp_path)

        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""paths = ["foo"]
target = "esp32"
build_dir = "build_@t"
junitxml = "test.xml"
keep_going = true
""")

        # test basic config
        with monkeypatch.context() as m:
            m.setenv('PATH', 'foo')  # let build fail
            m.setattr('sys.argv', ['idf-build-apps', 'build'])
            with pytest.raises(SystemExit):
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '1'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '0'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test cli overrides config
        with monkeypatch.context() as m:
            m.setenv('PATH', 'foo')  # let build fail
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--build-dir', 'build_hi_@t'])
            with pytest.raises(SystemExit):
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '1'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '0'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_hi_esp32'

        # test cli action_true
        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--dry-run'])
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test config store_true set to true
        with open(IDF_BUILD_APPS_TOML_FN, 'a') as fw:
            fw.write('\ndry_run = true\n')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build'])
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test config store_true set to false, but CLI set to true
        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""paths = ["foo"]
build_dir = "build_@t"
junitxml = "test.xml"
dry_run = false
""")

        with monkeypatch.context() as m:
            m.setattr(
                'sys.argv', ['idf-build-apps', 'build', '--default-build-targets', 'esp32', 'esp32s2', '--dry-run']
            )
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '2'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'
        assert test_suite.findall('testcase')[1].attrib['name'] == 'foo/build_esp32s2'

    def test_config_file_by_cli(self, tmp_path, monkeypatch):
        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write('paths = ["foo"]')

        with NamedTemporaryFile(mode='w', suffix='.toml') as ft:
            ft.write('paths = ["bar"]')
            ft.flush()

            with monkeypatch.context() as m:
                m.setattr(
                    'sys.argv',
                    [
                        'idf-build-apps',
                        'build',
                        '-t',
                        'esp32',
                        '--config-file',
                        ft.name,
                        '--junitxml',
                        'test.xml',
                        '--dry-run',
                    ],
                )
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'bar/build'



# SPDX-FileCopyrightText: 2024-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import enum
import glob
import inspect
import logging
import os
import re
import sys
import typing as t
from copy import deepcopy
from dataclasses import dataclass
from io import TextIOWrapper
from pathlib import Path
from typing import Any

from pydantic import AliasChoices, Field, computed_field, field_validator
from pydantic.fields import FieldInfo
from pydantic_core.core_schema import ValidationInfo
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from typing_extensions import Concatenate, ParamSpec

from . import SESSION_ARGS, App, setup_logging
from .constants import ALL_TARGETS, IDF_BUILD_APPS_TOML_FN, SUPPORTED_TARGETS
from .manifest.manifest import FolderRule, Manifest
from .utils import InvalidCommand, files_matches_patterns, semicolon_separated_str_to_list, to_absolute_path, to_list
from .vendors.pydantic_sources import PyprojectTomlConfigSettingsSource, TomlConfigSettingsSource

LOGGER = logging.getLogger(__name__)


class ValidateMethod(str, enum.Enum):
    TO_LIST = 'to_list'
    EXPAND_VARS = 'expand_vars'


@dataclass
class FieldMetadata:
    """
    dataclass field metadata. All fields are optional.
    Some fields are used in argparse while running :func:`add_args_to_parser`.

    :param validate_method: validate method for the field
    :param deprecates: deprecates field names, used in argparse
    :param shorthand: shorthand for the argument, used in argparse
    :param action: action for the argument, used in argparse
    :param nargs: nargs for the argument, used in argparse
    :param choices: choices for the argument, used in argparse
    :param type: type for the argument, used in argparse
    :param required: whether the argument is required, used in argparse
    :param default: default value for the argument, used in argparse
    :param hidden: whether the argument is hidden, used in argparse
    """

    # validate method
    validate_method: t.Optional[t.List[str]] = None
    # the field description will be copied from the deprecates field if not specified
    deprecates: t.Optional[t.Dict[str, t.Dict[str, t.Any]]] = None
    shorthand: t.Optional[str] = None
    # argparse_kwargs
    action: t.Optional[str] = None
    nargs: t.Optional[str] = None
    choices: t.Optional[t.List[str]] = None
    type: t.Optional[t.Callable] = None
    required: bool = False
    # usually default is not needed. only set it when different from the default value of the field
    default: t.Any = None
    # hidden field, use deprecated instead, or hide it in the argparse
    hidden: bool = False


P = ParamSpec('P')
T = t.TypeVar('T')


def _wrap_with_metadata(
    _: t.Callable[P, t.Any],
) -> t.Callable[[t.Callable[..., T]], t.Callable[Concatenate[t.Optional[FieldMetadata], P], T]]:
    """Patch the function signature with metadata args"""

    def return_func(func: t.Callable[..., T]) -> t.Callable[Concatenate[t.Optional[FieldMetadata], P], T]:
        return t.cast(t.Callable[Concatenate[t.Optional[FieldMetadata], P], T], func)

    return return_func


@_wrap_with_metadata(Field)
def field(meta: t.Optional[FieldMetadata], *args, **kwargs):
    """field with metadata"""
    f = Field(*args, **kwargs)
    f.metadata.append(meta)
    return f


def get_meta(f: FieldInfo) -> t.Optional[FieldMetadata]:
    """
    Get the metadata of the field

    :param f: field
    :return: metadata of the field if exists, None otherwise
    """
    for m in f.metadata:
        if isinstance(m, FieldMetadata):
            return m

    return None


class BaseArguments(BaseSettings):
    """Base settings class for all settings classes"""

    CONFIG_FILE_PATH: t.ClassVar[t.Optional[Path]] = None

    model_config = SettingsConfigDict(
        # these below two are supported in pydantic 2.6
        pyproject_toml_table_header=('tool', 'idf-build-apps'),
        pyproject_toml_depth=sys.maxsize,
        extra='ignore',  # we're supporting pydantic <2.6 as well, so we ignore extra fields
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: t.Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        dotenv_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        file_secret_settings: PydanticBaseSettingsSource,  # noqa: ARG003
    ) -> t.Tuple[PydanticBaseSettingsSource, ...]:
        sources: t.Tuple[PydanticBaseSettingsSource, ...] = (init_settings,)
        if cls.CONFIG_FILE_PATH is None:
            sources += (TomlConfigSettingsSource(settings_cls, toml_file=Path(IDF_BUILD_APPS_TOML_FN)),)
            sources += (PyprojectTomlConfigSettingsSource(settings_cls, toml_file=Path('pyproject.toml')),)
        else:
            sources += (TomlConfigSettingsSource(settings_cls, toml_file=Path(cls.CONFIG_FILE_PATH)),)
            sources += (PyprojectTomlConfigSettingsSource(settings_cls, toml_file=Path(cls.CONFIG_FILE_PATH)),)

        return sources

    @field_validator('*', mode='before')
    @classmethod
    def validate_by_validate_methods(cls, v: t.Any, info: ValidationInfo):
        if info.field_name and info.field_name in cls.model_fields:
            f = cls.model_fields[info.field_name]
            meta = get_meta(f)
            if meta and meta.validate_method:
                for method in meta.validate_method:
                    if method == ValidateMethod.TO_LIST:
                        v = to_list(v)
                    elif method == ValidateMethod.EXPAND_VARS:
                        v = os.path.expandvars(v)
                    else:
                        raise NotImplementedError(f'Unknown validate method: {method}')

                return v

        return v


class GlobalArguments(BaseArguments):
    verbose: int = field(
        FieldMetadata(
            shorthand='-v',
            action='count',
        ),
        description='Verbosity level. By default set to WARNING. Specify -v for INFO, -vv for DEBUG',
        default=0,  # type: ignore
    )
    log_file: t.Optional[str] = field(
        None,
        description='Path to the log file, if not specified logs will be printed to stderr',
        default=None,  # type: ignore
    )
    no_color: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Disable colored output',
        default=False,  # type: ignore
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        setup_logging(self.verbose, self.log_file, not self.no_color)


class DependencyDrivenBuildArguments(GlobalArguments):
    manifest_files: t.Optional[t.List[t.Union[Path, str]]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'manifest_file': {
                    'nargs': '+',
                },
            },
            nargs='+',
        ),
        description='Path to the manifest files which contains the build test rules of the apps',
        validation_alias=AliasChoices('manifest_files', 'manifest_file'),
        default=None,  # type: ignore
    )
    manifest_filepatterns: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            nargs='+',
        ),
        description='space-separated list of file patterns to search for the manifest files. '
        'The matched files will be loaded as the manifest files.',
        default=None,  # type: ignore
    )
    manifest_rootpath: str = field(
        FieldMetadata(
            validate_method=[ValidateMethod.EXPAND_VARS],
        ),
        description='Root path to resolve the relative paths defined in the manifest files. '
        'By default set to the current directory. Support environment variables.',
        default=os.curdir,  # type: ignore
    )
    modified_components: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            type=semicolon_separated_str_to_list,
        ),
        description='semicolon-separated list of modified components. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list.',
        default=None,  # type: ignore
    )
    modified_files: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            type=semicolon_separated_str_to_list,
        ),
        description='semicolon-separated list of modified files. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list.',
        default=None,  # type: ignore
    )
    deactivate_dependency_driven_build_by_components: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'ignore_app_dependencies_components': {
                    'type': semicolon_separated_str_to_list,
                    'shorthand': '-ic',
                }
            },
            type=semicolon_separated_str_to_list,
            shorthand='-dc',
        ),
        description='semicolon-separated list of components. '
        'dependency-driven build feature will be deactivated when any of these components are modified. '
        'Must be specified together with --modified-components. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list.',
        validation_alias=AliasChoices(
            'deactivate_dependency_driven_build_by_components', 'ignore_app_dependencies_components'
        ),
        default=None,  # type: ignore
    )
    deactivate_dependency_driven_build_by_filepatterns: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'ignore_app_dependencies_filepatterns': {
                    'type': semicolon_separated_str_to_list,
                    'shorthand': '-if',
                }
            },
            type=semicolon_separated_str_to_list,
            shorthand='-df',
        ),
        description='semicolon-separated list of file patterns. '
        'dependency-driven build feature will be deactivated when any of matched files are modified. '
        'Must be specified together with --modified-files. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list.',
        validation_alias=AliasChoices(
            'deactivate_dependency_driven_build_by_filepatterns', 'ignore_app_dependencies_filepatterns'
        ),
        default=None,  # type: ignore
    )
    check_manifest_rules: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Check if all folders defined in the manifest files exist. Fail if not',
        default=False,  # type: ignore
    )
    compare_manifest_sha_filepath: t.Optional[str] = field(
        None,
        description='Path to the file containing the hash of the manifest rules. '
        'Compare the hash with the current manifest rules. '
        'All matched apps will be built if the corresponding manifest rule is modified',
        default=None,  # type: ignore
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        if self.manifest_filepatterns:
            matched_paths = set()
            for pat in [to_absolute_path(p, self.manifest_rootpath) for p in self.manifest_filepatterns]:
                matched_paths.update(glob.glob(str(pat), recursive=True))

            if matched_paths:
                if self.manifest_files:
                    self.manifest_files.extend(matched_paths)
                else:
                    self.manifest_files = list(matched_paths)

        Manifest.CHECK_MANIFEST_RULES = self.check_manifest_rules
        if self.manifest_files:
            App.MANIFEST = Manifest.from_files(
                self.manifest_files,
                root_path=to_absolute_path(self.manifest_rootpath),
            )

    @property
    def dependency_driven_build_enabled(self) -> bool:
        """
        Check if the dependency-driven build feature is enabled

        :return: True if enabled, False otherwise
        """
        # not check since modified_components and modified_files are not passed
        if self.modified_components is None and self.modified_files is None:
            return False

        # not check since deactivate_dependency_driven_build_by_components is passed and matched
        if (
            self.deactivate_dependency_driven_build_by_components
            and self.modified_components is not None
            and set(self.modified_components).intersection(self.deactivate_dependency_driven_build_by_components)
        ):
            LOGGER.info(
                'Build all apps since modified components %s matches ignored components %s',
                ', '.join(self.modified_components),
                ', '.join(self.deactivate_dependency_driven_build_by_components),
            )
            return False

        # not check since deactivate_dependency_driven_build_by_filepatterns is passed and matched
        if (
            self.deactivate_dependency_driven_build_by_filepatterns
            and self.modified_files is not None
            and files_matches_patterns(
                self.modified_files, self.deactivate_dependency_driven_build_by_filepatterns, self.manifest_rootpath
            )
        ):
            LOGGER.info(
                'Build all apps since modified files %s matches ignored file patterns %s',
                ', '.join(self.modified_files),
                ', '.join(self.deactivate_dependency_driven_build_by_filepatterns),
            )
            return False

        return True

    @property
    def modified_manifest_rules_folders(self) -> t.Optional[t.Set[str]]:
        if self.compare_manifest_sha_filepath and App.MANIFEST is not None:
            return App.MANIFEST.diff_sha_with_filepath(self.compare_manifest_sha_filepath, use_abspath=True)

        return None


class FindBuildArguments(DependencyDrivenBuildArguments):
    paths: t.List[str] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            shorthand='-p',
            nargs='*',
        ),
        description='Paths to the directories containing the apps. By default set to the current directory',
        default=os.curdir,  # type: ignore
    )
    target: str = field(
        FieldMetadata(
            shorthand='-t',
        ),
        description='Filter the apps by target. By default set to "all"',
        default='all',  # type: ignore
    )
    build_system: t.Union[str, t.Type[App]] = field(
        FieldMetadata(
            choices=['cmake', 'make'],
        ),
        description='Filter the apps by build system. By default set to "cmake"',
        default='cmake',  # type: ignore
    )
    recursive: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Search for apps recursively under the specified paths',
        default=False,  # type: ignore
    )
    exclude: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            nargs='+',
        ),
        description='Ignore the specified directories while searching recursively',
        validation_alias=AliasChoices('exclude', 'exclude_list'),
        default=None,  # type: ignore
    )
    work_dir: t.Optional[str] = field(
        None,
        description='Copy the app to this directory before building. '
        'By default set to the app directory. Can expand placeholders',
        default=None,  # type: ignore
    )
    build_dir: str = field(
        None,
        description='Build directory for the app. By default set to "build". '
        'When set to relative path, it will be treated as relative to the app directory. '
        'Can expand placeholders',
        default='build',  # type: ignore
    )
    build_log_filename: t.Optional[str] = field(
        FieldMetadata(
            deprecates={'build_log': {}},
        ),
        description='Log filename under the build directory instead of stdout. Can expand placeholders',
        validation_alias=AliasChoices('build_log_filename', 'build_log'),
        default=None,  # type: ignore
    )
    size_json_filename: t.Optional[str] = field(
        FieldMetadata(
            deprecates={'size_file': {}},
        ),
        description='`idf.py size` output file under the build directory when specified. Can expand placeholders',
        validation_alias=AliasChoices('size_json_filename', 'size_file'),
        default=None,  # type: ignore
    )
    config_rules: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'config': {'nargs': '+'},
            },
            nargs='+',
        ),
        description='Defines the rules of building the project with pre-set sdkconfig files. '
        'Supports FILENAME[=NAME] or FILEPATTERN format. '
        'FILENAME is the filename of the sdkconfig file, relative to the app directory. '
        'Optional NAME is the name of the configuration. '
        'if not specified, the filename is used as the name. '
        'FILEPATTERN is the filename of the sdkconfig file with a single wildcard character (*). '
        'The NAME is the value matched by the wildcard',
        validation_alias=AliasChoices('config_rules', 'config_rules_str', 'config'),
        default=None,  # type: ignore
    )
    override_sdkconfig_items: t.Optional[str] = field(
        None,
        description='A comma-separated list of key=value pairs to override the sdkconfig items',
        default=None,  # type: ignore
    )
    override_sdkconfig_files: t.Optional[str] = field(
        None,
        description='A comma-separated list of sdkconfig files to override the sdkconfig items. '
        'When set to relative path, it will be treated as relative to the current directory',
        default=None,  # type: ignore
    )
    sdkconfig_defaults: t.Optional[str] = field(
        None,
        description='A semicolon-separated list of sdkconfig files passed to `idf.py -DSDKCONFIG_DEFAULTS`. '
        'SDKCONFIG_DEFAULTS environment variable is used when not specified',
        default=os.getenv('SDKCONFIG_DEFAULTS', None),  # type: ignore
    )
    check_warnings: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Check for warnings in the build output. Fail if any warnings are found',
        default=False,  # type: ignore
    )
    default_build_targets: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            nargs='+',
        ),
        description='space-separated list of the default enabled build targets for the apps. '
        'When not specified, the default value is the targets listed by `idf.py --list-targets`',
        default=None,  # type: ignore
    )
    enable_preview_targets: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='When enabled, the default build targets will be set to all apps, '
        'including the preview targets. As the targets defined in `idf.py --list-targets --preview`',
        default=False,  # type: ignore
    )
    disable_targets: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            nargs='+',
        ),
        description='space-separated list of targets that should be disabled to all apps.',
        default=None,  # type: ignore
    )
    include_skipped_apps: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Include the skipped apps in the output, together with the enabled ones',
        default=False,  # type: ignore
    )
    include_disabled_apps: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Include the disabled apps in the output, together with the enabled ones',
        default=False,  # type: ignore
    )
    include_all_apps: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Include skipped, and disabled apps in the output, together with the enabled ones',
        default=False,  # type: ignore
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        if not self.paths:
            raise InvalidCommand('At least one path must be specified')

        if not self.target:
            LOGGER.debug('--target is missing. Set --target as "all".')
            self.target = 'all'

        if self.default_build_targets:
            default_build_targets = []
            for target in self.default_build_targets:
                if target not in ALL_TARGETS:
                    LOGGER.warning(
                        f'Ignoring... Unrecognizable target {target} specified with "--default-build-targets". '
                        f'Current ESP-IDF available targets: {ALL_TARGETS}'
                    )
                elif target not in default_build_targets:
                    default_build_targets.append(target)
            self.default_build_targets = default_build_targets
            LOGGER.info('Overriding default build targets to %s', self.default_build_targets)
            FolderRule.DEFAULT_BUILD_TARGETS = self.default_build_targets
        elif self.enable_preview_targets:
            self.default_build_targets = deepcopy(ALL_TARGETS)
            LOGGER.info('Overriding default build targets to %s', self.default_build_targets)
            FolderRule.DEFAULT_BUILD_TARGETS = self.default_build_targets
        else:
            # restore default build targets
            FolderRule.DEFAULT_BUILD_TARGETS = SUPPORTED_TARGETS

        if self.disable_targets and FolderRule.DEFAULT_BUILD_TARGETS:
            LOGGER.info('Disable targets: %s', self.disable_targets)
            self.default_build_targets = [t for t in FolderRule.DEFAULT_BUILD_TARGETS if t not in self.disable_targets]
            FolderRule.DEFAULT_BUILD_TARGETS = self.default_build_targets

        if self.override_sdkconfig_files or self.override_sdkconfig_items:
            SESSION_ARGS.set(self)


class FindArguments(FindBuildArguments):
    output: t.Optional[str] = field(
        FieldMetadata(
            shorthand='-o',
        ),
        description='Record the found apps to the specified file instead of stdout',
        default=None,  # type: ignore
    )
    output_format: str = field(
        FieldMetadata(
            choices=['raw', 'json'],
        ),
        description='Output format of the found apps. '
        'In "raw" format, each line is a json string serialized from the app model. '
        'In "json" format, the output is a json list of the serialized app models',
        default='raw',  # type: ignore
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        if self.include_all_apps:
            self.include_skipped_apps = True
            self.include_disabled_apps = True

        if self.output and self.output.endswith('.json') and self.output_format in ['raw', None]:
            LOGGER.debug('Detecting output file ends with ".json", writing as json file.')
            self.output_format = 'json'


class BuildArguments(FindBuildArguments):
    build_verbose: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Enable verbose output of the build system',
        default=False,  # type: ignore
    )
    parallel_count: int = field(
        FieldMetadata(
            type=int,
        ),
        description='Number of parallel build jobs in total. '
        'Specified together with --parallel-index. '
        'The given apps will be divided into parallel_count parts, '
        'and the current run will build the parallel_index-th part',
        default=1,  # type: ignore
    )
    parallel_index: int = field(
        FieldMetadata(
            type=int,
        ),
        description='Index (1-based) of the parallel build job. '
        'Specified together with --parallel-count. '
        'The given apps will be divided into parallel_count parts, '
        'and the current run will build the parallel_index-th part',
        default=1,  # type: ignore
    )
    dry_run: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Skip the actual build, only print the build process',
        default=False,  # type: ignore
    )
    keep_going: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Continue building the next app when the current build fails',
        default=False,  # type: ignore
    )
    no_preserve: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Do not preserve the build directory after a successful build',
        default=False,  # type: ignore
    )
    ignore_warning_strs: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'ignore_warning_str': {'nargs': '+'},
            },
            nargs='+',
        ),
        description='space-separated list of patterns. Ignore the warnings in the build output that match the patterns',
        validation_alias=AliasChoices('ignore_warning_strs', 'ignore_warning_str'),
        default=None,  # type: ignore
    )
    ignore_warning_files: t.Optional[t.List[t.Union[str, TextIOWrapper]]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            deprecates={
                'ignore_warning_file': {
                    'type': argparse.FileType('r'),
                }
            },
            nargs='+',
            type=argparse.FileType('r'),
        ),
        description='Path to the files containing the patterns to ignore the warnings in the build output',
        validation_alias=AliasChoices('ignore_warning_files', 'ignore_warning_file'),
        default=None,  # type: ignore
    )
    copy_sdkconfig: bool = field(
        FieldMetadata(
            action='store_true',
        ),
        description='Copy the sdkconfig file to the build directory',
        default=False,  # type: ignore
    )

    # Attrs that support placeholders
    collect_size_info_filename: t.Optional[str] = field(
        FieldMetadata(
            deprecates={'collect_size_info': {}},
            hidden=True,
        ),
        description='Record size json filepath of the built apps to the specified file. '
        'Each line is a json string. Can expand placeholders @p',
        validation_alias=AliasChoices('collect_size_info_filename', 'collect_size_info'),
        default=None,  # type: ignore
        exclude=True,  # computed field is used
    )
    collect_app_info_filename: t.Optional[str] = field(
        FieldMetadata(
            deprecates={'collect_app_info': {}},
            hidden=True,
        ),
        description='Record serialized app model of the built apps to the specified file. '
        'Each line is a json string. Can expand placeholders @p',
        validation_alias=AliasChoices('collect_app_info_filename', 'collect_app_info'),
        default=None,  # type: ignore
        exclude=True,  # computed field is used
    )
    junitxml_filename: t.Optional[str] = field(
        FieldMetadata(
            deprecates={'junitxml': {}},
            hidden=True,
        ),
        description='Path to the junitxml file to record the build results. Can expand placeholder @p',
        validation_alias=AliasChoices('junitxml_filename', 'junitxml'),
        default=None,  # type: ignore
        exclude=True,  # computed field is used
    )
    # used for expanding placeholders
    PARALLEL_INDEX_PLACEHOLDER: t.ClassVar[str] = '@p'  # replace it with the parallel index

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        ignore_warnings_regexes = []
        if self.ignore_warning_strs:
            for s in self.ignore_warning_strs:
                ignore_warnings_regexes.append(re.compile(s))
        if self.ignore_warning_files:
            for f in self.ignore_warning_files:
                if isinstance(f, str):
                    with open(f) as fr:
                        for s in fr:
                            ignore_warnings_regexes.append(re.compile(s.strip()))
                else:
                    for s in f:
                        ignore_warnings_regexes.append(re.compile(s.strip()))
        App.IGNORE_WARNS_REGEXES = ignore_warnings_regexes

    @computed_field  # type: ignore
    @property
    def collect_size_info(self) -> t.Optional[str]:
        if self.collect_size_info_filename:
            return self.collect_size_info_filename.replace(self.PARALLEL_INDEX_PLACEHOLDER, str(self.parallel_index))

        return None

    @computed_field  # type: ignore
    @property
    def collect_app_info(self) -> t.Optional[str]:
        if self.collect_app_info_filename:
            return self.collect_app_info_filename.replace(self.PARALLEL_INDEX_PLACEHOLDER, str(self.parallel_index))

        return None

    @computed_field  # type: ignore
    @property
    def junitxml(self) -> t.Optional[str]:
        if self.junitxml_filename:
            return self.junitxml_filename.replace(self.PARALLEL_INDEX_PLACEHOLDER, str(self.parallel_index))

        return None


class DumpManifestShaArguments(GlobalArguments):
    manifest_files: t.Optional[t.List[str]] = field(
        FieldMetadata(
            validate_method=[ValidateMethod.TO_LIST],
            nargs='+',
            required=True,
        ),
        description='Path to the manifest files which contains the build test rules of the apps',
        default=None,  # type: ignore
    )

    output: t.Optional[str] = field(
        FieldMetadata(
            shorthand='-o',
            required=True,
        ),
        description='Path to the output file to record the sha256 hash of the manifest rules',
        default=None,  # type: ignore
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        if not self.manifest_files:
            raise InvalidCommand('Manifest files are required to dump the SHA values.')
        if not self.output:
            raise InvalidCommand('Output file is required to dump the SHA values.')


def _snake_case_to_cli_arg_name(s: str) -> str:
    return f'--{s.replace("_", "-")}'


def add_args_to_parser(argument_cls: t.Type[BaseArguments], parser: argparse.ArgumentParser) -> None:
    """
    Add arguments to the parser from the argument class.

    FieldMetadata is used to set the argparse options.

    :param argument_cls: argument class
    :param parser: argparse parser
    """
    for f_name, f in argument_cls.model_fields.items():
        f_meta = get_meta(f)
        if f_meta and f_meta.deprecates:
            for dep_f_name, dep_f_kwargs in f_meta.deprecates.items():
                _names = [_snake_case_to_cli_arg_name(dep_f_name)]
                _shorthand = dep_f_kwargs.pop('shorthand', None)
                if _shorthand:
                    _names.append(_shorthand)

                if f_meta.hidden:  # f is hidden, use deprecated field instead
                    help_msg = f.description
                else:
                    help_msg = f'[Deprecated] Use {_snake_case_to_cli_arg_name(f_name)} instead'

                parser.add_argument(
                    *_names,
                    **dep_f_kwargs,
                    help=help_msg,
                )

        if f_meta and f_meta.hidden:
            continue

        names = [_snake_case_to_cli_arg_name(f_name)]
        if f_meta and f_meta.shorthand:
            names.append(f_meta.shorthand)

        kwargs: t.Dict[str, t.Any] = {}
        if f_meta:
            if f_meta.type:
                kwargs['type'] = f_meta.type
            if f_meta.required:
                kwargs['required'] = True
            if f_meta.action:
                kwargs['action'] = f_meta.action
                # to make the CLI override config file work
                if f_meta.action == 'store_true':
                    kwargs['default'] = None

            if f_meta.nargs:
                kwargs['nargs'] = f_meta.nargs
            if f_meta.choices:
                kwargs['choices'] = f_meta.choices
            if f_meta.default:
                kwargs['default'] = f_meta.default

        # here in CLI arguments, don't set the default to field.default
        # otherwise it will override the config file settings

        parser.add_argument(
            *names,
            **kwargs,
            help=f.description,
        )


def add_args_to_obj_doc_as_params(argument_cls: t.Type[GlobalArguments], obj: t.Any = None) -> None:
    """
    Add arguments to the function as parameters.

    :param argument_cls: argument class
    :param obj: object to add the docstring to
    """
    _obj = obj or argument_cls
    _doc_str = _obj.__doc__ or ''
    _doc_str += '\n'

    for f_name, f in argument_cls.model_fields.items():
        # typing generic alias is not a class
        _annotation = f.annotation.__name__ if inspect.isclass(f.annotation) else f.annotation
        _doc_str += f'    :param {f_name}: {f.description}\n'
        _doc_str += f'    :type {f_name}: {_annotation}\n'

    _obj.__doc__ = _doc_str


def apply_config_file(config_file: t.Optional[str] = None, reset: bool = False) -> None:
    def _subclasses(klass: t.Type[T]) -> t.Set[t.Type[T]]:
        return set(klass.__subclasses__()).union([s for c in klass.__subclasses__() for s in _subclasses(c)])

    if reset:
        BaseArguments.CONFIG_FILE_PATH = None
        for cls in _subclasses(BaseArguments):
            cls.CONFIG_FILE_PATH = None

    if config_file:
        if os.path.isfile(config_file):
            p = Path(config_file)
            BaseArguments.CONFIG_FILE_PATH = p
            for cls in _subclasses(BaseArguments):
                cls.CONFIG_FILE_PATH = p
        else:
            LOGGER.warning(f'Config file {config_file} does not exist. Ignoring...')



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

import idf_build_apps
from idf_build_apps import (
    App,
    setup_logging,
)
from idf_build_apps.args import apply_config_file
from idf_build_apps.constants import SUPPORTED_TARGETS
from idf_build_apps.manifest.manifest import FolderRule


@pytest.fixture(autouse=True)
def clean_cls_attr(tmp_path):
    App.MANIFEST = None
    FolderRule.DEFAULT_BUILD_TARGETS = SUPPORTED_TARGETS
    idf_build_apps.SESSION_ARGS.clean()
    apply_config_file(reset=True)
    os.chdir(tmp_path)


@pytest.fixture(autouse=True)
def setup_logging_debug():
    setup_logging(1)


def create_project(name, folder):
    p = str(folder / name)
    os.makedirs(p)
    os.makedirs(os.path.join(p, 'main'))

    with open(os.path.join(p, 'CMakeLists.txt'), 'w') as fw:
        fw.write(
            f"""cmake_minimum_required(VERSION 3.16)
include($ENV{{IDF_PATH}}/tools/cmake/project.cmake)
project({name})
"""
        )

    with open(os.path.join(p, 'main', 'CMakeLists.txt'), 'w') as fw:
        fw.write(
            f"""idf_component_register(SRCS "{name}.c"
INCLUDE_DIRS ".")
"""
        )

    with open(os.path.join(p, 'main', f'{name}.c'), 'w') as fw:
        fw.write(
            """#include <stdio.h>
void app_main(void) {}
"""
        )


@pytest.fixture
def sha_of_enable_only_esp32():
    sha = FolderRule('test1', enable=[{'if': 'IDF_TARGET == "esp32"'}]).sha

    # !!! ONLY CHANGE IT WHEN NECESSARY !!!
    assert (
        sha
        == 'bfc1c61176cfb76169eab6c4f00dbcc4d7886fee4b93ede5fac2c005dd85db640464e9b03986d3da3bfaa4d109b053862c07dc4d5a407e58f773a8f710ec60cb'  # noqa: E501
    )

    return sha


@pytest.fixture
def sha_of_enable_esp32_or_esp32s2():
    sha = FolderRule('test1', enable=[{'if': 'IDF_TARGET == "esp32" or IDF_TARGET == "esp32s2"'}]).sha

    # !!! ONLY CHANGE IT WHEN NECESSARY !!!
    assert (
        sha
        == '9ab121a0d39bcb590465837091e82dfd798cd1ff9579e92c23e8bebaee127b46751108f0de3953993cb7993903e45d78851fc465d774a606b0ab1251fbe4b9f5'  # noqa: E501
    )

    return sha



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import shutil
import subprocess

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from datetime import datetime

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'idf-build-apps'
project_homepage = 'https://github.com/espressif/idf-build-apps'
copyright = f'2023-{datetime.now().year}, Espressif Systems (Shanghai) Co., Ltd.'  # noqa: A001
author = 'Fu Hanxi'
languages = ['en']
version = '2.x'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_copybutton',
    'myst_parser',
    'sphinxcontrib.mermaid',
    'sphinxarg.ext',
    'sphinx_tabs.tabs',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_css_files = ['theme_overrides.css']
html_logo = '../_static/espressif-logo.svg'
html_static_path = ['../_static']
html_theme = 'sphinx_rtd_theme'

# mermaid 10.2.0 will show syntax error
# use fixed version instead
mermaid_version = '10.6.1'

autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'show-inheritance': True,
    'exclude-members': 'model_computed_fields,model_config,model_fields,model_post_init',
}


def generate_api_docs(language):
    from idf_build_apps.args import (
        BuildArguments,
        FindArguments,
        add_args_to_obj_doc_as_params,
    )
    from idf_build_apps.main import build_apps, find_apps

    docs_dir = os.path.dirname(__file__)
    api_dir = os.path.join(docs_dir, language, 'references', 'api')
    if os.path.isdir(api_dir):
        shutil.rmtree(api_dir)

    # --- MOCK DOCSTRINGS By Arguments ---
    add_args_to_obj_doc_as_params(FindArguments)
    add_args_to_obj_doc_as_params(BuildArguments)
    add_args_to_obj_doc_as_params(FindArguments, find_apps)
    add_args_to_obj_doc_as_params(BuildArguments, build_apps)
    # --- MOCK DOCSTRINGS FINISHED ---

    subprocess.run(
        [
            'sphinx-apidoc',
            os.path.join(docs_dir, '..', 'idf_build_apps'),
            '-f',
            '-H',
            'API Reference',
            '--no-headings',
            '-t',
            '_apidoc_templates',
            '-o',
            api_dir,
        ]
    )



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import platform
import re
import socket
import sys
import typing as t


def get_size(b: float) -> str:
    for unit in ['', 'K', 'M', 'G', 'T', 'P']:
        if b < 1024:
            return f'{b:.2f}{unit}B'
        b /= 1024

    return f'{b:.2f}EB'


def get_processor_name():
    if platform.processor():
        return platform.processor()

    # read from /proc/cpuinfo
    if os.path.isfile('/proc/cpuinfo'):
        try:
            with open('/proc/cpuinfo') as f:
                for line in f:
                    if 'model name' in line:
                        return re.sub('.*model name.*:', '', line, count=1).strip()
        except Exception:
            pass

    return ''


def get_sys_info() -> t.Dict[str, str]:
    info = {
        'platform': platform.system(),
        'platform-release': platform.release(),
        'architecture': platform.machine(),
        'hostname': socket.gethostname(),
        'processor': get_processor_name(),
        'cpu_count': str(os.cpu_count()) if os.cpu_count() else 'Unknown',
    }

    if sys.platform != 'win32':
        info['ram'] = get_size(os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES'))

    return info



# SPDX-FileCopyrightText: 2022 Samuel Colvin and other contributors
# SPDX-License-Identifier: MIT

"""
Partially copied from https://github.com/pydantic/pydantic-settings v2.5.2
since python 3.7 version got dropped at pydantic-settings 2.1.0
but the feature we need introduced in 2.2.0

For contributing history please refer to the original github page
For the full license text refer to
https://github.com/pydantic/pydantic-settings/blob/9b73e924cab136d876907af0c6836dcca09ac35c/LICENSE

Modifications:
- use toml instead of tomli when python < 3.11
- stop using global variables
- fix some warnings
- recursively find TOML file.
"""

import logging
import os
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic_settings import InitSettingsSource
from pydantic_settings.main import BaseSettings

from idf_build_apps.constants import IDF_BUILD_APPS_TOML_FN

PathType = Union[Path, str, List[Union[Path, str]], Tuple[Union[Path, str], ...]]
DEFAULT_PATH = Path('')
LOGGER = logging.getLogger(__name__)


class ConfigFileSourceMixin(ABC):
    def _read_files(self, files: Optional[PathType]) -> Dict[str, Any]:
        if files is None:
            return {}
        if isinstance(files, (str, os.PathLike)):
            files = [files]
        kwargs: Dict[str, Any] = {}
        for file in files:
            file_path = Path(file).expanduser()
            if file_path.is_file():
                kwargs.update(self._read_file(file_path))
        return kwargs

    @abstractmethod
    def _read_file(self, path: Optional[Path]) -> Dict[str, Any]:
        pass


class TomlConfigSettingsSource(InitSettingsSource, ConfigFileSourceMixin):
    """
    A source class that loads variables from a TOML file
    """

    def __init__(
        self,
        settings_cls: Type[BaseSettings],
        toml_file: Optional[Path] = DEFAULT_PATH,
    ):
        self.toml_file_path = self._pick_toml_file(
            toml_file,
            settings_cls.model_config.get('pyproject_toml_depth', sys.maxsize),
            IDF_BUILD_APPS_TOML_FN,
        )
        self.toml_data = self._read_files(self.toml_file_path)
        super().__init__(settings_cls, self.toml_data)

    def _read_file(self, path: Optional[Path]) -> Dict[str, Any]:
        if not path or not path.is_file():
            return {}

        if sys.version_info < (3, 11):
            import toml

            with open(path) as toml_file:
                return toml.load(toml_file)
        else:
            import tomllib

            with open(path, 'rb') as toml_file:
                return tomllib.load(toml_file)

    @staticmethod
    def _pick_toml_file(provided: Optional[Path], depth: int, filename: str) -> Optional[Path]:
        """
        Pick a file path to use. If a file path is provided, use it. Otherwise, search up the directory tree for a
        file with the given name.

        :param provided: Explicit path provided when instantiating this class.
        :param depth: Number of directories up the tree to check of a pyproject.toml.
        """
        if provided and Path(provided).is_file():
            fp = provided.resolve()
            LOGGER.debug(f'Loading config file: {fp}')
            return fp

        rv = Path.cwd()
        count = -1
        while count < depth:
            if len(rv.parts) == 1:
                break

            fp = rv / filename
            if fp.is_file():
                LOGGER.debug(f'Loading config file: {fp}')
                return fp

            rv = rv.parent
            count += 1

        return None


class PyprojectTomlConfigSettingsSource(TomlConfigSettingsSource):
    """
    A source class that loads variables from a `pyproject.toml` file.
    """

    def __init__(
        self,
        settings_cls: Type[BaseSettings],
        toml_file: Optional[Path] = None,
    ) -> None:
        self.toml_file_path = self._pick_toml_file(
            toml_file,
            settings_cls.model_config.get('pyproject_toml_depth', sys.maxsize),
            'pyproject.toml',
        )
        self.toml_table_header: Tuple[str, ...] = settings_cls.model_config.get(
            'pyproject_toml_table_header',
            ('tool', 'idf-build-apps'),
        )
        self.toml_data = self._read_files(self.toml_file_path)
        for key in self.toml_table_header:
            self.toml_data = self.toml_data.get(key, {})
        super(TomlConfigSettingsSource, self).__init__(settings_cls, self.toml_data)



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import enum
import os

import esp_bool_parser

IDF_PATH = esp_bool_parser.IDF_PATH
IDF_PY = os.path.join(IDF_PATH, 'tools', 'idf.py')
IDF_SIZE_PY = os.path.join(IDF_PATH, 'tools', 'idf_size.py')

PROJECT_DESCRIPTION_JSON = 'project_description.json'
DEFAULT_SDKCONFIG = 'sdkconfig.defaults'

SUPPORTED_TARGETS = esp_bool_parser.SUPPORTED_TARGETS
PREVIEW_TARGETS = esp_bool_parser.PREVIEW_TARGETS
ALL_TARGETS = esp_bool_parser.ALL_TARGETS
IDF_VERSION_MAJOR = esp_bool_parser.IDF_VERSION_MAJOR
IDF_VERSION_MINOR = esp_bool_parser.IDF_VERSION_MINOR
IDF_VERSION_PATCH = esp_bool_parser.IDF_VERSION_PATCH
IDF_VERSION = esp_bool_parser.IDF_VERSION


class BuildStatus(str, enum.Enum):
    UNKNOWN = 'unknown'
    DISABLED = 'disabled'
    SKIPPED = 'skipped'
    SHOULD_BE_BUILT = 'should be built'
    FAILED = 'build failed'
    SUCCESS = 'build success'


completion_instructions = """
With the `--activate` option, detect your shell type and add the appropriate commands to your shell's config file
so that it runs on startup. You will likely have to restart.
or re-login for the autocompletion to start working.

You can also specify your shell using the `--shell` option.

If you do not want automatic modification of your shell configuration file
You can manually add the commands provided below to activate autocompletion.
or run them in your current terminal session for one-time activation.

Once again, you will likely have to restart
or re-login for the autocompletion to start working.

bash:
    eval "$(register-python-argcomplete idf-build-apps)"

zsh:
    To activate completions in zsh, first make sure compinit is marked for
    autoload and run autoload:

    autoload -U compinit
    compinit

    Afterwards you can enable completions for idf-build-apps:

    eval "$(register-python-argcomplete idf-build-apps)"

fish:
    # Not required to be in the config file, only run once
    register-python-argcomplete --shell fish idf-build-apps >~/.config/fish/completions/idf-build-apps.fish
"""
IDF_BUILD_APPS_TOML_FN = '.idf_build_apps.toml'



# PYTHON_ARGCOMPLETE_OK

# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import logging
import os
import sys
import textwrap
import typing as t
from pathlib import Path

import argcomplete
from pydantic import (
    Field,
    create_model,
)

from idf_build_apps.args import (
    BuildArguments,
    DumpManifestShaArguments,
    FindArguments,
    add_args_to_parser,
    apply_config_file,
)

from .app import (
    App,
    AppDeserializer,
    CMakeApp,
    MakeApp,
)
from .autocompletions import activate_completions
from .constants import ALL_TARGETS, BuildStatus, completion_instructions
from .finder import (
    _find_apps,
)
from .junit import (
    TestCase,
    TestReport,
    TestSuite,
)
from .manifest.manifest import (
    Manifest,
)
from .utils import (
    AutocompleteActivationError,
    InvalidCommand,
    drop_none_kwargs,
    get_parallel_start_stop,
    to_list,
)

LOGGER = logging.getLogger(__name__)


def find_apps(
    paths: t.Union[t.List[str], str, None] = None,
    target: t.Optional[str] = None,
    *,
    find_arguments: t.Optional[FindArguments] = None,
    config_file: t.Optional[str] = None,
    **kwargs,
) -> t.List[App]:
    """
    Find apps in the given paths for the specified target. For all kwargs, please refer to `FindArguments`

    :return: list of found apps
    """
    if config_file:
        apply_config_file(config_file)

    # compatible with old usage
    ## `preserve`
    if 'preserve' in kwargs:
        LOGGER.warning(
            'DEPRECATED: Passing "preserve" directly is deprecated. '
            'Pass "no_preserve" instead to disable preserving the build directory'
        )
        kwargs['no_preserve'] = not kwargs.pop('preserve')

    if find_arguments is None:
        find_arguments = FindArguments(
            paths=to_list(paths),  # type: ignore
            target=target,  # type: ignore
            **kwargs,
        )

    app_cls: t.Type[App]
    if isinstance(find_arguments.build_system, str):
        # backwards compatible
        if find_arguments.build_system == 'cmake':
            app_cls = CMakeApp
        elif find_arguments.build_system == 'make':
            app_cls = MakeApp
        else:
            raise ValueError('Only Support "make" and "cmake"')
    else:
        app_cls = find_arguments.build_system

    apps: t.Set[App] = set()
    if find_arguments.target == 'all':
        targets = ALL_TARGETS
        LOGGER.info('Searching for apps by all targets')
    else:
        targets = [find_arguments.target]
        LOGGER.info('Searching for apps by target: %s', find_arguments.target)

    for _t in targets:
        for _p in find_arguments.paths:
            LOGGER.debug('Searching for apps in path %s for target %s', _p, _t)
            apps.update(
                _find_apps(
                    _p,
                    _t,
                    app_cls=app_cls,
                    args=find_arguments,
                )
            )

    LOGGER.info('Found %d apps in total', len(apps))

    return sorted(apps)


def build_apps(
    apps: t.Union[t.List[App], App, None] = None,
    *,
    build_arguments: t.Optional[BuildArguments] = None,
    config_file: t.Optional[str] = None,
    **kwargs,
) -> int:
    """
    Build all the specified apps. For all kwargs, please refer to `BuildArguments`

    :return: exit code
    """
    if config_file:
        apply_config_file(config_file)

    # compatible with old usage
    ## `check_app_dependencies`
    if 'check_app_dependencies' in kwargs:
        LOGGER.warning(
            'DEPRECATED: Passing "check_app_dependencies" directly is deprecated. '
            'Pass "modified_components" instead to enable dependency-driven build feature'
        )
        kwargs.pop('check_app_dependencies')

    if build_arguments is None:
        build_arguments = BuildArguments(
            **kwargs,
        )

    apps = to_list(apps)
    if apps is None:
        apps = find_apps(
            find_arguments=FindArguments(
                **{k: v for k, v in build_arguments.model_dump().items() if k in FindArguments.model_fields}
            )
        )

    test_suite = TestSuite('build_apps')

    start, stop = get_parallel_start_stop(len(apps), build_arguments.parallel_count, build_arguments.parallel_index)
    LOGGER.info('Processing %d total apps: building apps %d-%d', len(apps), start, stop)

    # cleanup collect files if exists at this early-stage
    for f in (build_arguments.collect_app_info, build_arguments.collect_size_info, build_arguments.junitxml):
        if f and os.path.isfile(f):
            LOGGER.debug('Removing existing collect file: %s', f)
            os.remove(f)

    exit_code = 0

    # create empty files, avoid no file when no app is built
    if build_arguments.collect_app_info:
        LOGGER.debug('Creating empty app info file: %s', build_arguments.collect_app_info)
        Path(build_arguments.collect_app_info).touch()
    if build_arguments.collect_size_info:
        LOGGER.debug('Creating empty size info file: %s', build_arguments.collect_size_info)
        Path(build_arguments.collect_size_info).touch()

    for i, app in enumerate(apps):
        index = i + 1  # we use 1-based
        if index < start or index > stop:
            continue

        # attrs
        app.dry_run = build_arguments.dry_run
        app.index = index
        app.verbose = build_arguments.build_verbose
        app.copy_sdkconfig = build_arguments.copy_sdkconfig

        LOGGER.info('(%d/%d) Building app: %s', index, len(apps), app)

        app.build(
            manifest_rootpath=build_arguments.manifest_rootpath,
            modified_components=build_arguments.modified_components,
            modified_files=build_arguments.modified_files,
            check_app_dependencies=build_arguments.dependency_driven_build_enabled,
        )
        test_suite.add_test_case(TestCase.from_app(app))

        if app.build_comment:
            LOGGER.info('%s (%s)', app.build_status.value, app.build_comment)
        else:
            LOGGER.info('%s', app.build_status.value)

        if build_arguments.collect_app_info:
            with open(build_arguments.collect_app_info, 'a') as fw:
                fw.write(app.to_json() + '\n')
            LOGGER.debug('Recorded app info in file: %s', build_arguments.collect_app_info)

        if app.build_status == BuildStatus.FAILED:
            if not build_arguments.keep_going:
                LOGGER.error('Build failed and keep_going=False, stopping build process')
                return 1
            else:
                LOGGER.warning('Build failed but keep_going=True, continuing with next app')
                exit_code = 1
        elif app.build_status == BuildStatus.SUCCESS:
            if build_arguments.collect_size_info and app.size_json_path:
                if os.path.isfile(app.size_json_path):
                    with open(build_arguments.collect_size_info, 'a') as fw:
                        fw.write(
                            json.dumps(
                                {
                                    'app_name': app.name,
                                    'config_name': app.config_name,
                                    'target': app.target,
                                    'path': app.size_json_path,
                                }
                            )
                            + '\n'
                        )
                    LOGGER.debug('Recorded size info file path: %s', build_arguments.collect_size_info)

        LOGGER.info('')  # add one empty line for separating different builds

    if build_arguments.junitxml:
        TestReport([test_suite], build_arguments.junitxml).create_test_report()

    return exit_code


class IdfBuildAppsCliFormatter(argparse.HelpFormatter):
    LINE_SEP = '$LINE_SEP$'

    def _split_lines(self, text, width):
        parts = text.split(self.LINE_SEP)

        text = self._whitespace_matcher.sub(' ', parts[0]).strip()
        return textwrap.wrap(text, width) + parts[1:]

    def _get_help_string(self, action):
        """
        Add the default value to the option help message.

        ArgumentDefaultsHelpFormatter and BooleanOptionalAction when it isn't
        already present. This code will do that, detecting corner cases to
        prevent duplicates or cases where it wouldn't make sense to the end
        user.
        """
        _help = action.help
        if _help is None:
            _help = ''

        if action.dest == 'config_file':
            return _help

        if action.default is not argparse.SUPPRESS:
            if action.default is None:
                default_type = str
            else:
                default_type = type(action.default)

            if action.nargs in [argparse.ZERO_OR_MORE, argparse.ONE_OR_MORE]:
                _type = f'list[{default_type.__name__}]'
            else:
                _type = default_type.__name__

            defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                _help += f'{self.LINE_SEP} - default: %(default)s'

            _help += f'{self.LINE_SEP} - config name: {action.dest}'
            _help += f'{self.LINE_SEP} - config type: {_type}'

        return _help


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='Tools for building ESP-IDF related apps. '
        'Some CLI options can be expanded by the following placeholders, like "--work-dir", "--build-dir", etc.:\n'
        '- @t: would be replaced by the target chip type\n'
        '- @w: would be replaced by the wildcard, usually the sdkconfig\n'
        '- @n: would be replaced by the app name\n'
        '- @f: would be replaced by the escaped app path (replaced "/" to "_")\n'
        '- @v: Would be replaced by the ESP-IDF version like `5_3_0`\n'
        '- @i: would be replaced by the build index (only available in `build` command)\n'
        '- @p: would be replaced by the parallel index (only available in `build` command)',
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    actions = parser.add_subparsers(dest='action', required=True)

    common_args = argparse.ArgumentParser(add_help=False)
    common_args.add_argument('--config-file', help='Path to the config file')

    ########
    # Find #
    ########
    find_parser = actions.add_parser(
        'find',
        help='Find the buildable applications. Run `idf-build-apps find --help` for more information on a command.',
        description='Find the buildable applications in the given path or paths for specified chips. '
        '`--path` and `--target` options must be provided. '
        'By default, print the found apps in stdout. '
        'To find apps for all chips use the `--target` option with the `all` argument.',
        formatter_class=IdfBuildAppsCliFormatter,
        parents=[common_args],
    )
    add_args_to_parser(FindArguments, find_parser)

    #########
    # Build #
    #########
    build_parser = actions.add_parser(
        'build',
        help='Build the found applications. Run `idf-build-apps build --help` for more information on a command.',
        description='Build the application in the given path or paths for specified chips. '
        '`--path` and `--target` options must be provided.',
        formatter_class=IdfBuildAppsCliFormatter,
        parents=[common_args],
    )
    add_args_to_parser(BuildArguments, build_parser)

    ###############
    # Completions #
    ###############
    completions_parser = actions.add_parser(
        'completions',
        help='Add the autocompletion activation script to the shell rc file. '
        'Run `idf-build-apps completions --help` for more information on a command.',
        description='Without `--activate` option print instructions for manual activation. '
        'With the `--activate` option, add the autocompletion activation script to the shell rc file '
        'for bash, zsh, or fish. Other shells are not supported.'
        'The `--shell` option is used only with the `--activate` option, '
        'if provided, add the autocompletion activation script to the given shell; '
        'without this argument, will detect shell type automatically. '
        'May need to restart or re-login for the autocompletion to start working',
        formatter_class=IdfBuildAppsCliFormatter,
    )
    completions_parser.add_argument(
        '-a', '--activate', action='store_true', help='Activate autocompletion automatically and permanently. '
    )
    completions_parser.add_argument(
        '-s',
        '--shell',
        choices=['bash', 'zsh', 'fish'],
        help='Specify the shell type for the autocomplete activation script.',
    )

    ############################
    # Dump Manifest SHA Values #
    ############################
    dump_manifest_parser = actions.add_parser(
        'dump-manifest-sha',
        help='Dump the manifest files SHA values. '
        'This could be useful in CI to check if the manifest files are changed.',
        parents=[common_args],
    )
    add_args_to_parser(DumpManifestShaArguments, dump_manifest_parser)

    return parser


def handle_completions(args: argparse.Namespace) -> None:
    if args.activate:
        if not args.shell:
            args.shell = 'auto'
        activate_completions(args.shell)
    elif not args.activate and args.shell:
        raise AutocompleteActivationError('The --shell option can only be used with the --activate option.')
    else:
        print(completion_instructions)


def main():
    parser = get_parser()
    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    if args.action == 'completions':
        handle_completions(args)
        sys.exit(0)

    kwargs = vars(args)
    action = kwargs.pop('action')
    config_file = kwargs.pop('config_file')
    if config_file:
        apply_config_file(config_file)

    if action == 'dump-manifest-sha':
        arguments = DumpManifestShaArguments(**drop_none_kwargs(kwargs))
        Manifest.from_files(arguments.manifest_files).dump_sha_values(arguments.output)
        sys.exit(0)
    elif action == 'find':
        arguments = FindArguments(**drop_none_kwargs(kwargs))
    else:
        arguments = BuildArguments(**drop_none_kwargs(kwargs))

    # real call starts here
    # build also needs to find first
    apps = find_apps(args.paths, args.target, find_arguments=arguments)
    if isinstance(arguments, FindArguments):  # find only
        if arguments.output:
            os.makedirs(os.path.dirname(os.path.realpath(arguments.output)), exist_ok=True)
            with open(arguments.output, 'w') as fw:
                if arguments.output_format == 'raw':
                    for app in apps:
                        fw.write(app.to_json() + '\n')
                elif arguments.output_format == 'json':
                    fw.write(json.dumps([app.model_dump() for app in apps], indent=2))
                else:
                    raise InvalidCommand(f'Output format {arguments.output_format} is not supported.')
        else:
            for app in apps:
                print(app)

        sys.exit(0)

    # build
    if arguments.no_preserve:
        for app in apps:
            app.preserve = False

    ret_code = build_apps(apps, build_arguments=arguments)

    built_apps = [app for app in apps if app.build_status == BuildStatus.SUCCESS]
    if built_apps:
        print('Successfully built the following apps:')
        for app in built_apps:
            print(f'  {app}')

    skipped_apps = [app for app in apps if app.build_status == BuildStatus.SKIPPED]
    if skipped_apps:
        print('Skipped building the following apps:')
        for app in skipped_apps:
            print(f'  {app}')

    failed_apps = [app for app in apps if app.build_status == BuildStatus.FAILED]
    if failed_apps:
        print('Failed building the following apps:')
        for app in failed_apps:
            print(f'  {app}')

    if ret_code != 0:
        sys.exit(ret_code)


def json_to_app(json_str: str, extra_classes: t.Optional[t.List[t.Type[App]]] = None) -> App:
    """
    Deserialize json string to App object

    .. note::

        You can pass extra_cls to support custom App class. A custom App class must be a subclass of App, and have a
        different value of `build_system`. For example, a custom CMake app

        .. code:: python

           class CustomApp(CMakeApp):
               build_system: Literal['custom_cmake'] = 'custom_cmake'

        Then you can pass the :class:`CustomApp` class to the :attr:`extra_cls` argument

        .. code:: python

           json_str = CustomApp('.', 'esp32').to_json()
           json_to_app(json_str, extra_classes=[CustomApp])

    :param json_str: json string
    :param extra_classes: extra App class
    :return: App object
    """
    types = [App, CMakeApp, MakeApp]
    if extra_classes:
        types.extend(extra_classes)

    custom_deserializer = create_model(
        '_CustomDeserializer',
        app=(t.Union[tuple(types)], Field(discriminator='build_system')),
        __base__=AppDeserializer,
    )

    return custom_deserializer.from_json(json_str)



# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import sys

language = 'en'

sys.path.insert(0, os.path.abspath('../'))

from conf_common import *  # noqa

generate_api_docs(language)  # noqa



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import copy
import functools
import json
import logging
import os
import re
import shutil
import sys
import typing as t
from datetime import datetime, timezone
from pathlib import (
    Path,
)

from packaging.version import (
    Version,
)
from pydantic import (
    Field,
    computed_field,
)

from . import (
    SESSION_ARGS,
)
from .constants import (
    DEFAULT_SDKCONFIG,
    IDF_PY,
    IDF_SIZE_PY,
    IDF_VERSION,
    IDF_VERSION_MAJOR,
    IDF_VERSION_MINOR,
    IDF_VERSION_PATCH,
    PREVIEW_TARGETS,
    PROJECT_DESCRIPTION_JSON,
    BuildStatus,
)
from .manifest.manifest import (
    FolderRule,
    Manifest,
)
from .utils import (
    BaseModel,
    BuildError,
    Literal,
    files_matches_patterns,
    find_first_match,
    rmdir,
    subprocess_run,
    to_absolute_path,
    to_list,
)

LOGGER = logging.getLogger(__name__)


class App(BaseModel):
    TARGET_PLACEHOLDER: t.ClassVar[str] = '@t'  # replace it with self.target
    WILDCARD_PLACEHOLDER: t.ClassVar[str] = '@w'  # replace it with the wildcard, usually the sdkconfig
    NAME_PLACEHOLDER: t.ClassVar[str] = '@n'  # replace it with self.name
    FULL_NAME_PLACEHOLDER: t.ClassVar[str] = '@f'  # replace it with escaped self.app_dir
    IDF_VERSION_PLACEHOLDER: t.ClassVar[str] = '@v'  # replace it with the IDF version
    INDEX_PLACEHOLDER: t.ClassVar[str] = '@i'  # replace it with the build index (while build_apps)

    SDKCONFIG_LINE_REGEX: t.ClassVar[t.Pattern] = re.compile(r'^([^=]+)=\"?([^\"\n]*)\"?\n*$')

    # could be assigned later, used for filtering out apps by supported_targets
    MANIFEST: t.ClassVar[t.Optional[Manifest]] = None
    # This RE will match GCC errors and many other fatal build errors and warnings as well
    LOG_ERROR_WARNING_REGEX: t.ClassVar[t.Pattern] = re.compile(r'(?:error|warning):', re.MULTILINE | re.IGNORECASE)
    # Log this many trailing lines from a failed build log, also
    LOG_DEBUG_LINES: t.ClassVar[int] = 25
    # IGNORE_WARNING_REGEX is a regex for warnings to be ignored. Could be assigned later
    IGNORE_WARNS_REGEXES: t.ClassVar[t.List[t.Pattern]] = []

    # ------------------
    # Instance variables
    # ------------------
    build_system: Literal['unknown'] = 'unknown'

    app_dir: str
    target: str
    sdkconfig_path: t.Optional[str] = None
    config_name: t.Optional[str] = None
    sdkconfig_defaults_str: t.Optional[str] = None
    # all parameters
    _kwargs: t.Dict[str, t.Any]

    # Attrs that support placeholders
    _work_dir: t.Optional[str] = None
    _build_dir: t.Optional[str] = None

    _build_log_filename: t.Optional[str] = None
    _size_json_filename: t.Optional[str] = None

    dry_run: bool = False
    verbose: bool = False
    check_warnings: bool = False
    preserve: bool = True
    copy_sdkconfig: bool = False

    # build_apps() related
    index: t.Optional[int] = None

    # build status related
    build_status: BuildStatus = BuildStatus.UNKNOWN
    build_comment: t.Optional[str] = None

    _build_duration: float = 0
    _build_timestamp: t.Optional[datetime] = None

    __EQ_IGNORE_FIELDS__ = [
        'build_comment',
    ]
    __EQ_TUNE_FIELDS__ = {
        'app_dir': lambda x: (os.path.realpath(os.path.expanduser(x))),
        'work_dir': lambda x: (os.path.realpath(os.path.expanduser(x))),
        'build_dir': lambda x: (os.path.realpath(os.path.expanduser(x))),
    }

    def __init__(
        self,
        app_dir: str,
        target: str,
        *,
        work_dir: t.Optional[str] = None,
        build_dir: str = 'build',
        build_log_filename: t.Optional[str] = None,
        size_json_filename: t.Optional[str] = None,
        **kwargs: t.Any,
    ) -> None:
        kwargs.update(
            {
                'app_dir': app_dir,
                'target': target,
            }
        )
        super().__init__(**kwargs)

        # These internal variables store the paths with environment variables and placeholders;
        # Public properties with similar names use the _expand method to get the actual paths.
        self._work_dir = work_dir or app_dir
        self._build_dir = build_dir or 'build'

        self._build_log_filename = build_log_filename
        self._size_json_filename = size_json_filename
        self._is_build_log_path_temp = not bool(build_log_filename)

        # pass all parameters to initialize hook method
        kwargs.update(
            {
                'work_dir': self._work_dir,
                'build_dir': self._build_dir,
                'build_log_filename': build_log_filename,
                'size_json_filename': size_json_filename,
            }
        )
        self._kwargs = kwargs
        self._initialize_hook(**kwargs)

        # private attrs, won't be dumped to json
        self._checked_should_build = False

        self._sdkconfig_files, self._sdkconfig_files_defined_target = self._process_sdkconfig_files()

    @classmethod
    def from_another(cls, other: 'App', **kwargs) -> 'App':
        """Init New App from another, with different parameters.

        eg: new_app = [Cmake]App.from_another(app, config='debug')
        """
        new_kwargs = copy.deepcopy(other._kwargs)
        new_kwargs.update(kwargs)
        return cls(**new_kwargs)

    def _initialize_hook(self, **kwargs):
        """
        Called after variables initialized, before actions such as creating logger.
        """
        pass

    def __str__(self):
        default_fmt = '({}) App {}, target {}, sdkconfig {}, build in {}'
        default_args = [
            self.build_system,
            self.app_dir,
            self.target,
            self.sdkconfig_path or '(default)',
            self.build_path,
        ]

        if self.build_status in (BuildStatus.UNKNOWN, BuildStatus.SHOULD_BE_BUILT):
            return default_fmt.format(*default_args)

        default_fmt += ', {} in {}s'
        default_args += [
            self.build_status.value,
            self._build_duration,
        ]

        if self.build_comment:
            default_fmt += ': {}'
            default_args.append(self.build_comment)

        return default_fmt.format(*default_args)

    @property
    def sdkconfig_defaults_candidates(self) -> t.List[str]:
        if self.sdkconfig_defaults_str is not None:
            return self.sdkconfig_defaults_str.split(';')

        if os.getenv('SDKCONFIG_DEFAULTS', None) is not None:
            return os.getenv('SDKCONFIG_DEFAULTS', '').split(';')

        return [DEFAULT_SDKCONFIG]

    @t.overload
    def _expand(self, path: None) -> None: ...

    @t.overload
    def _expand(self, path: str) -> str: ...

    def _expand(self, path):
        """
        Internal method, expands any of the placeholders in {app,work,build} paths.
        """
        if not path:
            return path

        if self.index is not None:
            path = path.replace(self.INDEX_PLACEHOLDER, str(self.index))
        path = path.replace(
            self.IDF_VERSION_PLACEHOLDER, f'{IDF_VERSION_MAJOR}_{IDF_VERSION_MINOR}_{IDF_VERSION_PATCH}'
        )
        path = path.replace(self.TARGET_PLACEHOLDER, self.target)
        path = path.replace(self.NAME_PLACEHOLDER, self.name)
        if self.FULL_NAME_PLACEHOLDER in path:  # to avoid recursion to the call to app_dir in the next line:
            path = path.replace(self.FULL_NAME_PLACEHOLDER, self.app_dir.replace(os.path.sep, '_'))
        wildcard_pos = path.find(self.WILDCARD_PLACEHOLDER)
        if wildcard_pos != -1:
            if self.config_name:
                # if config name is defined, put it in place of the placeholder
                path = path.replace(self.WILDCARD_PLACEHOLDER, self.config_name)
            else:
                # otherwise, remove the placeholder and one character on the left
                # (which is usually an underscore, dash, or other delimiter)
                left_of_wildcard = max(0, wildcard_pos - 1)
                right_of_wildcard = wildcard_pos + len(self.WILDCARD_PLACEHOLDER)
                path = path[0:left_of_wildcard] + path[right_of_wildcard:]
        path = os.path.expandvars(path)
        return path

    @property
    def name(self) -> str:
        base_name = os.path.basename(self.app_dir)
        # '.' for relative path like '.'
        # '' for path endswith '/'
        if base_name in ['.', '']:
            return os.path.basename(os.path.abspath(self.app_dir))
        return base_name

    @computed_field  # type: ignore
    @property
    def work_dir(self) -> str:
        """
        :return: directory where the app should be copied to, prior to the build.
        """
        return self._expand(self._work_dir)  # type: ignore

    @computed_field  # type: ignore
    @property
    def build_dir(self) -> str:
        """
        :return: build directory, either relative to the work directory (if relative path is used) or absolute path.
        """
        return self._expand(self._build_dir)  # type: ignore

    @property
    def build_path(self) -> str:
        if os.path.isabs(self.build_dir):
            return self.build_dir

        return os.path.join(self.work_dir, self.build_dir)

    @computed_field  # type: ignore
    @property
    def build_log_filename(self) -> t.Optional[str]:
        return self._expand(self._build_log_filename)

    @property
    def build_log_path(self) -> str:
        if self.build_log_filename:
            return os.path.join(self.build_path, self.build_log_filename)

        # use a temp file if build log path is not specified
        return os.path.join(self.build_path, '.temp.build.log')

    @computed_field  # type: ignore
    @property
    def size_json_filename(self) -> t.Optional[str]:
        if self.target == 'linux':
            # esp-idf-size does not support linux target
            return None

        return self._expand(self._size_json_filename)

    @property
    def size_json_path(self) -> t.Optional[str]:
        if self.size_json_filename:
            return os.path.join(self.build_path, self.size_json_filename)

        return None

    def _process_sdkconfig_files(self) -> t.Tuple[t.List[str], t.Optional[str]]:
        """
        Expand environment variables in default sdkconfig files and remove some CI related settings.
        """
        real_sdkconfig_files: t.List[str] = []
        sdkconfig_files_defined_target: t.Optional[str] = None

        # put the expanded variable files in a temporary directory
        # will remove if the content is the same as the original one
        expanded_dir = os.path.join(self.work_dir, 'expanded_sdkconfig_files', os.path.basename(self.build_dir))
        if not os.path.isdir(expanded_dir):
            os.makedirs(expanded_dir)

        for f in self.sdkconfig_defaults_candidates + ([self.sdkconfig_path] if self.sdkconfig_path else []):
            # use filepath if abs/rel already point to itself
            if not os.path.isfile(f):
                # find it in the app_dir
                LOGGER.debug('sdkconfig file %s not found, checking under app_dir...', f)
                f = os.path.join(self.app_dir, f)
                if not os.path.isfile(f):
                    LOGGER.debug('sdkconfig file %s not found, skipping...', f)
                    continue

            expanded_fp = os.path.join(expanded_dir, os.path.basename(f))
            with open(f) as fr:
                with open(expanded_fp, 'w') as fw:
                    for line in fr:
                        line = os.path.expandvars(line)

                        m = self.SDKCONFIG_LINE_REGEX.match(line)
                        if m:
                            key = m.group(1)
                            if key == 'CONFIG_IDF_TARGET':
                                sdkconfig_files_defined_target = m.group(2)

                            if isinstance(self, CMakeApp):
                                if key in self.SDKCONFIG_TEST_OPTS:
                                    self.cmake_vars[key] = m.group(2)
                                    continue

                                if key in self.SDKCONFIG_IGNORE_OPTS:
                                    continue

                        fw.write(line)

            with open(f) as fr:
                with open(expanded_fp) as new_fr:
                    if fr.read() == new_fr.read():
                        LOGGER.debug('Use sdkconfig file %s', f)
                        try:
                            os.unlink(expanded_fp)
                        except OSError:
                            LOGGER.debug('Failed to remove file %s', expanded_fp)
                        real_sdkconfig_files.append(f)
                    else:
                        LOGGER.debug('Expand sdkconfig file %s to %s', f, expanded_fp)
                        real_sdkconfig_files.append(expanded_fp)
                        # copy the related target-specific sdkconfig files
                        par_dir = os.path.abspath(os.path.join(f, '..'))
                        for target_specific_file in (
                            os.path.join(par_dir, str(p))
                            for p in Path(par_dir).glob(os.path.basename(f) + f'.{self.target}')
                        ):
                            LOGGER.debug(
                                'Copy target-specific sdkconfig file %s to %s', target_specific_file, expanded_dir
                            )
                            shutil.copy(target_specific_file, expanded_dir)

        # remove if expanded folder is empty
        try:
            os.rmdir(expanded_dir)
        except OSError:
            pass

        try:
            os.rmdir(os.path.join(self.work_dir, 'expanded_sdkconfig_files'))
        except OSError:
            pass

        if SESSION_ARGS.override_sdkconfig_file_path:
            real_sdkconfig_files.append(SESSION_ARGS.override_sdkconfig_file_path)
            if 'CONFIG_IDF_TARGET' in SESSION_ARGS.override_sdkconfig_items:
                sdkconfig_files_defined_target = SESSION_ARGS.override_sdkconfig_items['CONFIG_IDF_TARGET']

        return real_sdkconfig_files, sdkconfig_files_defined_target

    @property
    def sdkconfig_files_defined_idf_target(self) -> t.Optional[str]:
        return self._sdkconfig_files_defined_target

    @property
    def sdkconfig_files(self) -> t.List[str]:
        return [os.path.abspath(file) for file in self._sdkconfig_files]

    @property
    def depends_components(self) -> t.List[str]:
        if self.MANIFEST:
            return self.MANIFEST.depends_components(
                self.app_dir, self.sdkconfig_files_defined_idf_target, self.config_name
            )

        return []

    @property
    def depends_filepatterns(self) -> t.List[str]:
        if self.MANIFEST:
            return self.MANIFEST.depends_filepatterns(
                self.app_dir, self.sdkconfig_files_defined_idf_target, self.config_name
            )

        return []

    @property
    def supported_targets(self) -> t.List[str]:
        if self.MANIFEST:
            return self.MANIFEST.enable_build_targets(
                self.app_dir, self.sdkconfig_files_defined_idf_target, self.config_name
            )

        if self.sdkconfig_files_defined_idf_target:
            return [self.sdkconfig_files_defined_idf_target]

        return FolderRule.DEFAULT_BUILD_TARGETS

    @property
    def verified_targets(self) -> t.List[str]:
        if self.MANIFEST:
            return self.MANIFEST.enable_test_targets(
                self.app_dir, self.sdkconfig_files_defined_idf_target, self.config_name
            )

        return []

    def record_build_duration(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self._build_timestamp = datetime.now(timezone.utc)
            try:
                return func(self, *args, **kwargs)
            finally:
                self._build_duration = (datetime.now(timezone.utc) - self._build_timestamp).total_seconds()

        return wrapper

    def _pre_build(self) -> None:
        if self.build_status == BuildStatus.SKIPPED:
            return

        if self.work_dir != self.app_dir:
            if os.path.exists(self.work_dir):
                LOGGER.debug('Removed existing work dir: %s', self.work_dir)
                if not self.dry_run:
                    shutil.rmtree(self.work_dir)

            LOGGER.debug('Copied app from %s to %s', self.app_dir, self.work_dir)
            if not self.dry_run:
                # if the new directory inside the original directory,
                # make sure not to go into recursion.
                ignore = shutil.ignore_patterns(
                    os.path.basename(self.work_dir),
                    # also ignore files which may be present in the work directory
                    'build',
                    'sdkconfig',
                )

                shutil.copytree(self.app_dir, self.work_dir, ignore=ignore, symlinks=True)

        if os.path.exists(self.build_path):
            LOGGER.debug('Removed existing build dir: %s', self.build_path)
            if not self.dry_run:
                shutil.rmtree(self.build_path)

        if not self.dry_run:
            os.makedirs(self.build_path, exist_ok=True)

        sdkconfig_file = os.path.join(self.work_dir, 'sdkconfig')
        if os.path.exists(sdkconfig_file):
            LOGGER.debug('Removed existing sdkconfig file: %s', sdkconfig_file)
            if not self.dry_run:
                os.unlink(sdkconfig_file)

        if os.path.isfile(self.build_log_path):
            LOGGER.debug('Removed existing build log file: %s', self.build_log_path)
            if not self.dry_run:
                os.unlink(self.build_log_path)
        elif not self.dry_run:
            os.makedirs(os.path.dirname(self.build_log_path), exist_ok=True)
        LOGGER.info('Writing build log to %s', self.build_log_path)

        if self.dry_run:
            self.build_status = BuildStatus.SKIPPED
            self.build_comment = 'dry run'
            return

    @record_build_duration  # type: ignore
    def build(
        self,
        *,
        manifest_rootpath: t.Optional[str] = None,
        modified_components: t.Union[t.List[str], str, None] = None,
        modified_files: t.Union[t.List[str], str, None] = None,
        check_app_dependencies: bool = False,
    ) -> None:
        if self.build_status not in (
            BuildStatus.UNKNOWN,
            BuildStatus.SHOULD_BE_BUILT,
        ):
            self.build_comment = f'Build {self.build_status.value}. Skipping...'
            return

        # real build starts here
        self._pre_build()

        try:
            self._build(
                manifest_rootpath=manifest_rootpath,
                modified_components=to_list(modified_components),
                modified_files=to_list(modified_files),
                check_app_dependencies=check_app_dependencies,
            )
        except BuildError as e:
            self.build_status = BuildStatus.FAILED
            self.build_comment = str(e)

        self._post_build()

        self._finalize()

    def _post_build(self) -> None:
        """Post build actions for failed/success builds"""
        if self.build_status not in (
            BuildStatus.FAILED,
            BuildStatus.SUCCESS,
        ):
            return

        # both status applied
        if self.copy_sdkconfig:
            try:
                shutil.copy(
                    os.path.join(self.work_dir, 'sdkconfig'),
                    os.path.join(self.build_path, 'sdkconfig'),
                )
            except Exception as e:
                LOGGER.warning('Copy sdkconfig file from failed: %s', e)
            else:
                LOGGER.debug('Copied sdkconfig file from %s to %s', self.work_dir, self.build_path)

        # for originally success builds generate size.json if enabled
        #
        # for the rest of the actions, we need to check if there's further build warnings
        # to tell if this build is successful or not
        if self.build_status == BuildStatus.SUCCESS:
            self.write_size_json()

        if not os.path.isfile(self.build_log_path):
            LOGGER.warning(f'{self.build_log_path} does not exist. Skipping post build actions...')
            return

        # check warnings
        has_unignored_warning = False
        with open(self.build_log_path) as fr:
            lines = [line.rstrip() for line in fr if line.rstrip()]
            for line in lines:
                is_error_or_warning, ignored = self.is_error_or_warning(line)
                if is_error_or_warning:
                    if ignored:
                        LOGGER.info('[Ignored warning] %s', line)
                    else:
                        LOGGER.warning('%s', line)
                        has_unignored_warning = True

        # for failed builds, print last few lines to help debug
        if self.build_status == BuildStatus.FAILED:
            # print last few lines to help debug
            LOGGER.error(
                'Last %s lines from the build log "%s":',
                self.LOG_DEBUG_LINES,
                self.build_log_path,
            )
            for line in lines[-self.LOG_DEBUG_LINES :]:
                LOGGER.error('%s', line)
        # correct build status for originally successful builds
        elif self.build_status == BuildStatus.SUCCESS:
            if self.check_warnings and has_unignored_warning:
                self.build_status = BuildStatus.FAILED
                self.build_comment = 'build succeeded with warnings'
            elif has_unignored_warning:
                self.build_comment = 'build succeeded with warnings'

    def _finalize(self) -> None:
        """
        Actions for real success builds
        """
        if self.build_status != BuildStatus.SUCCESS:
            return

        # remove temp log file
        if self._is_build_log_path_temp:
            os.unlink(self.build_log_path)
            LOGGER.debug('Removed success build temporary log file: %s', self.build_log_path)

        # Cleanup build directory if not preserving
        if not self.preserve:
            exclude_list = []
            if self.size_json_path:
                exclude_list.append(os.path.basename(self.size_json_path))
            exclude_list.append(os.path.basename(self.build_log_path))

            rmdir(
                self.build_path,
                exclude_file_patterns=exclude_list,
            )
            LOGGER.debug('Removed built binaries under: %s', self.build_path)

    def _build(
        self,
        *,
        manifest_rootpath: t.Optional[str] = None,
        modified_components: t.Optional[t.List[str]] = None,
        modified_files: t.Optional[t.List[str]] = None,
        check_app_dependencies: bool = False,
    ) -> None:
        pass

    def _write_size_json(self) -> None:
        if not self.size_json_path:
            return

        map_file = find_first_match('*.map', self.build_path)
        if not map_file:
            LOGGER.warning(
                '.map file not found. Cannot write size json to file: %s',
                self.size_json_path,
            )
            return

        if IDF_VERSION >= Version('4.1'):
            subprocess_run(
                [
                    sys.executable,
                    str(IDF_SIZE_PY),
                ]
                + (['--json'] if IDF_VERSION < Version('5.1') else ['--format', 'json'])
                + [
                    '-o',
                    self.size_json_path,
                    map_file,
                ],
                check=True,
            )
        else:
            with open(self.size_json_path, 'w') as fw:
                subprocess_run(
                    (
                        [
                            sys.executable,
                            str(IDF_SIZE_PY),
                            '--json',
                            map_file,
                        ]
                    ),
                    log_terminal=False,
                    log_fs=fw,
                    check=True,
                )

        LOGGER.debug('Generated size info to %s', self.size_json_path)

    def write_size_json(self) -> None:
        if self.target in PREVIEW_TARGETS:
            # targets in preview may not yet have support in esp-idf-size
            LOGGER.info('Skipping generation of size json for target %s as it is in preview.', self.target)
            return

        try:
            self._write_size_json()
        except Exception as e:
            LOGGER.warning('Failed to generate size json: %s', e)

    def to_json(self) -> str:
        return self.model_dump_json()

    def is_error_or_warning(self, line: str) -> t.Tuple[bool, bool]:
        if not self.LOG_ERROR_WARNING_REGEX.search(line):
            return False, False

        is_ignored = False
        for ignored in self.IGNORE_WARNS_REGEXES:
            if re.search(ignored, line):
                is_ignored = True
                break

        return True, is_ignored

    @classmethod
    def is_app(cls, path: str) -> bool:
        raise NotImplementedError('Please implement this function in sub classes')

    def is_modified(self, modified_files: t.Optional[t.List[str]]) -> bool:
        _app_dir_fullpath = to_absolute_path(self.app_dir)
        if modified_files:
            for f in modified_files:
                _f_fullpath = to_absolute_path(f)
                if os.path.basename(_f_fullpath).endswith('.md'):
                    continue

                if _f_fullpath.startswith(_app_dir_fullpath):
                    return True

        return False

    def check_should_build(
        self,
        *,
        manifest_rootpath: t.Optional[str] = None,
        modified_manifest_rules_folders: t.Optional[t.Set[str]] = None,
        check_app_dependencies: bool = False,
        modified_components: t.Optional[t.List[str]] = None,
        modified_files: t.Optional[t.List[str]] = None,
    ) -> None:
        if self.build_status != BuildStatus.UNKNOWN:
            return

        if not check_app_dependencies:
            self.build_status = BuildStatus.SHOULD_BE_BUILT
            self._checked_should_build = True
            return

        if (
            self.MANIFEST
            and modified_manifest_rules_folders
            and self.MANIFEST.most_suitable_rule(self.app_dir).folder in modified_manifest_rules_folders
        ):
            self.build_status = BuildStatus.SHOULD_BE_BUILT
            self.build_comment = 'current build modifies the related manifest rules'
            self._checked_should_build = True
            return

        if self.is_modified(modified_files):
            self.build_status = BuildStatus.SHOULD_BE_BUILT
            self.build_comment = 'current build modifies this app'
            self._checked_should_build = True
            return

        # if didn't modify any components, and no `depends_filepatterns` defined, skip
        if modified_components == [] and not self.depends_filepatterns:
            self.build_status = BuildStatus.SKIPPED
            self.build_comment = 'current build does not modify any components'
            self._checked_should_build = True
            return

        # if no special rules defined, we left it unknown and decide with idf.py reconfigure
        if not self.depends_components and not self.depends_filepatterns:
            # keep unknown
            self._checked_should_build = True
            self.build_comment = 'no special rules defined, run idf.py reconfigure to decide'
            return

        # check app dependencies
        modified_components = to_list(modified_components)
        modified_files = to_list(modified_files)

        # depends components?
        if self.depends_components and modified_components is not None:
            if set(self.depends_components).intersection(set(modified_components)):
                self._checked_should_build = True
                self.build_status = BuildStatus.SHOULD_BE_BUILT
                self.build_comment = (
                    f'Requires components: {", ".join(self.depends_components)}. '
                    f'Modified components: {", ".join(modified_components)}'
                )
                return

        # or depends file patterns?
        if self.depends_filepatterns and modified_files is not None:
            if files_matches_patterns(modified_files, self.depends_filepatterns, manifest_rootpath):
                self._checked_should_build = True
                self.build_status = BuildStatus.SHOULD_BE_BUILT
                self.build_comment = (
                    f'Requires file patterns: {", ".join(self.depends_filepatterns)}. '
                    f'Modified files: {", ".join(modified_files)}'
                )
                return

        # special rules defined, but not matched
        self.build_status = BuildStatus.SKIPPED
        self.build_comment = 'current build does not modify any components or files required by this app'
        self._checked_should_build = True


class MakeApp(App):
    MAKE_PROJECT_LINE: t.ClassVar[str] = r'include $(IDF_PATH)/make/project.mk'

    build_system: Literal['make'] = 'make'  # type: ignore

    @property
    def supported_targets(self) -> t.List[str]:
        if self.MANIFEST:
            return self.MANIFEST.enable_build_targets(
                self.app_dir, self.sdkconfig_files_defined_idf_target, self.config_name
            )

        if self.sdkconfig_files_defined_idf_target:
            return [self.sdkconfig_files_defined_idf_target]

        return ['esp8266', *FolderRule.DEFAULT_BUILD_TARGETS]

    def _build(
        self,
        *,
        manifest_rootpath: t.Optional[str] = None,
        modified_components: t.Optional[t.List[str]] = None,
        modified_files: t.Optional[t.List[str]] = None,
        check_app_dependencies: bool = False,
    ) -> None:
        super()._build(
            manifest_rootpath=manifest_rootpath,
            modified_components=modified_components,
            modified_files=modified_files,
            check_app_dependencies=check_app_dependencies,
        )

        # additional env variables
        additional_env_dict = {
            'IDF_TARGET': self.target,
            'BUILD_DIR_BASE': self.build_path,
        }

        commands = [
            # generate sdkconfig
            ['make', 'defconfig'],
            # build
            ['make', f'-j{os.cpu_count() or 1}'],
        ]

        for cmd in commands:
            subprocess_run(
                cmd,
                log_terminal=self._is_build_log_path_temp,
                log_fs=self.build_log_path,
                check=True,
                additional_env_dict=additional_env_dict,
                cwd=self.work_dir,
            )

        self.build_status = BuildStatus.SUCCESS

    @classmethod
    def is_app(cls, path: str) -> bool:
        makefile_path = os.path.join(path, 'Makefile')
        if not os.path.exists(makefile_path):
            return False

        with open(makefile_path) as makefile:
            makefile_content = makefile.read()

        if cls.MAKE_PROJECT_LINE not in makefile_content:
            return False

        return True


class CMakeApp(App):
    # If these keys are present in sdkconfig.defaults, they will be extracted and passed to CMake
    SDKCONFIG_TEST_OPTS: t.ClassVar[t.List[str]] = [
        'EXCLUDE_COMPONENTS',
        'TEST_EXCLUDE_COMPONENTS',
        'TEST_COMPONENTS',
    ]

    # These keys in sdkconfig.defaults are not propagated to the final sdkconfig file:
    SDKCONFIG_IGNORE_OPTS: t.ClassVar[t.List[str]] = ['TEST_GROUPS']

    # While ESP-IDF component CMakeLists files can be identified by the presence of 'idf_component_register' string,
    # there is no equivalent for the project CMakeLists files. This seems to be the best option...
    CMAKE_PROJECT_LINE: t.ClassVar[str] = r'include($ENV{IDF_PATH}/tools/cmake/project.cmake)'

    build_system: Literal['cmake'] = 'cmake'  # type: ignore

    cmake_vars: t.Dict[str, str] = {}

    def _build(
        self,
        *,
        manifest_rootpath: t.Optional[str] = None,
        modified_components: t.Optional[t.List[str]] = None,
        modified_files: t.Optional[t.List[str]] = None,
        check_app_dependencies: bool = False,
    ) -> None:
        super()._build(
            manifest_rootpath=manifest_rootpath,
            modified_components=modified_components,
            modified_files=modified_files,
            check_app_dependencies=check_app_dependencies,
        )

        if not self._checked_should_build:
            self.check_should_build(
                manifest_rootpath=manifest_rootpath,
                modified_components=modified_components,
                modified_files=modified_files,
                check_app_dependencies=check_app_dependencies,
            )

        # additional env variables
        # IDF_TARGET to bypass the idf.py build check
        additional_env_dict = {
            'IDF_TARGET': self.target,
        }

        # check if this app depends on components according to the project_description.json 'build_component' field.
        # the file is generated by `idf.py reconfigure`.
        common_args = [
            sys.executable,
            str(IDF_PY),
            '-B',
            self.build_path,
            '-C',
            self.work_dir,
            f'-DIDF_TARGET={self.target}',
            # set to ";" to disable `default` when no such variable
            '-DSDKCONFIG_DEFAULTS={}'.format(';'.join(self.sdkconfig_files) if self.sdkconfig_files else ';'),
        ]

        if self.build_status == BuildStatus.UNKNOWN and modified_components is not None and check_app_dependencies:
            subprocess_run(
                [*common_args, 'reconfigure'],
                log_terminal=self._is_build_log_path_temp,
                log_fs=self.build_log_path,
                check=True,
                additional_env_dict=additional_env_dict,
            )
            LOGGER.debug('generated project_description.json to check app dependencies')

            with open(os.path.join(self.build_path, PROJECT_DESCRIPTION_JSON)) as fr:
                build_components = {item for item in json.load(fr)['build_components'] if item}

            if not set(modified_components).intersection(set(build_components)):
                self.build_status = BuildStatus.SKIPPED
                self.build_comment = (
                    f'app {self.app_dir} depends components: {build_components}, '
                    f'while current build modified components: {modified_components}'
                )
                return

        if self.build_status == BuildStatus.SKIPPED:
            return

        # idf.py build
        if self.cmake_vars:
            for key, val in self.cmake_vars.items():
                common_args.append(f'-D{key}={val}')
            if 'TEST_EXCLUDE_COMPONENTS' in self.cmake_vars and 'TEST_COMPONENTS' not in self.cmake_vars:
                common_args.append('-DTESTS_ALL=1')
            if 'CONFIG_APP_BUILD_BOOTLOADER' in self.cmake_vars:
                # In case if secure_boot is enabled then for bootloader build need to add `bootloader` cmd
                common_args.append('bootloader')
        common_args.append('build')
        if self.verbose:
            common_args.append('-v')

        subprocess_run(
            common_args,
            log_terminal=self._is_build_log_path_temp,
            log_fs=self.build_log_path,
            check=True,
            additional_env_dict=additional_env_dict,
        )

        self.build_status = BuildStatus.SUCCESS

    @classmethod
    def is_app(cls, path: str) -> bool:
        cmakelists_path = os.path.join(path, 'CMakeLists.txt')
        if not os.path.exists(cmakelists_path):
            return False

        with open(cmakelists_path) as fr:
            cmakelists_file_content = fr.read()

        if not cmakelists_file_content:
            return False

        if cls.CMAKE_PROJECT_LINE not in cmakelists_file_content:
            return False

        return True


class AppDeserializer(BaseModel):
    app: t.Union[App, CMakeApp, MakeApp] = Field(discriminator='build_system')

    @classmethod
    def from_json(cls, json_data: t.Union[str, bytes, bytearray]) -> App:
        json_dict = json.loads(json_data.strip())
        return cls.model_validate({'app': json_dict}).app



# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Optional

from .utils import AutocompleteActivationError


def append_to_file(file_path: str, content: str) -> None:
    """Add commands to shell configuration file

    :param file_path: path to shell configurations file
    :param content: commands to add
    """
    if os.path.exists(file_path):
        with open(file_path) as file:
            if content.strip() in file.read():
                print(f'Autocompletion already set up in {file_path}')
                return
    with open(file_path, 'a') as file:
        file.write(f'\n# Begin added by idf-build-apps \n{content} \n# End added by idf-build-apps')
    print(f'Autocompletion added to {file_path}')


def activate_completions(shell_type: Optional[str]) -> None:
    """Activates autocompletion for supported shells.

    :raises AutocompleteActivationError: if the $SHELL env variable is empty, or if the detected shell is unsupported.
    """
    supported_shells = ['bash', 'zsh', 'fish']

    if shell_type == 'auto':
        shell_type = os.path.basename(os.environ.get('SHELL', ''))

    if not shell_type:
        raise AutocompleteActivationError('$SHELL is empty. Please provide your shell type with the `--shell` option')

    if shell_type not in supported_shells:
        raise AutocompleteActivationError('Unsupported shell. Autocompletion is supported for bash, zsh and fish.')

    if shell_type == 'bash':
        completion_command = 'eval "$(register-python-argcomplete idf-build-apps)"'
    elif shell_type == 'zsh':
        completion_command = (
            'autoload -U bashcompinit && bashcompinit && eval "$(register-python-argcomplete idf-build-apps)"'
        )
    elif shell_type == 'fish':
        completion_command = 'register-python-argcomplete --shell fish idf-build-apps | source'

    rc_file = {'bash': '~/.bashrc', 'zsh': '~/.zshrc', 'fish': '~/.config/fish/completions/idf-build-apps.fish'}

    shell_rc = os.path.expanduser(rc_file[shell_type])
    append_to_file(shell_rc, completion_command)



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
import yaml

import idf_build_apps
from idf_build_apps import setup_logging
from idf_build_apps.constants import (
    SUPPORTED_TARGETS,
)
from idf_build_apps.manifest.manifest import (
    IfClause,
    Manifest,
)
from idf_build_apps.utils import (
    InvalidIfClause,
    InvalidManifest,
)
from idf_build_apps.yaml import (
    parse,
)


def test_manifest_from_file_warning(tmp_path, capsys, monkeypatch):
    setup_logging()

    yaml_file = tmp_path / 'test.yml'
    yaml_file.write_text(
        """
test1:
    enable:
        - if: IDF_TARGET == "esp32" or IDF_TARGET == "esp32c3"
        - if: IDF_TARGET == "esp32s2" and IDF_VERSION_MAJOR in [4, 5]
    disable_test:
        - if: IDF_TARGET == "esp32c3"

test2:
    enable:
        - if: INCLUDE_DEFAULT == 0 and IDF_TARGET == "linux"
""",
        encoding='utf8',
    )

    manifest = Manifest.from_file(yaml_file, root_path=tmp_path)
    captured_err = capsys.readouterr().err.splitlines()
    msg_fmt = 'Folder "{}" does not exist. Please check your manifest file {}'
    # two warnings warn test1 test2 not exists
    assert len(captured_err) == 2
    assert msg_fmt.format(os.path.join(tmp_path, 'test1'), yaml_file) in captured_err[0]
    assert msg_fmt.format(os.path.join(tmp_path, 'test2'), yaml_file) in captured_err[1]

    assert manifest.enable_build_targets('test1') == ['esp32', 'esp32c3', 'esp32s2']
    assert manifest.enable_test_targets('test1') == ['esp32', 'esp32s2']
    assert manifest.enable_build_targets('test2') == ['linux']
    assert manifest.enable_test_targets('test2') == ['linux']

    monkeypatch.setattr(idf_build_apps.manifest.manifest.Manifest, 'CHECK_MANIFEST_RULES', True)
    with pytest.raises(InvalidManifest, match=msg_fmt.format(os.path.join(tmp_path, 'test1'), yaml_file)):
        Manifest.from_file(yaml_file, root_path=tmp_path)

    # test with folder that has the same prefix as one of the folders in the manifest
    assert manifest.enable_build_targets('test23') == sorted(SUPPORTED_TARGETS)


def test_repr(tmp_path):
    yaml_file = tmp_path / 'test.yml'

    yaml_file.write_text(
        """
test1:
  enable:
    - if: IDF_TARGET == "esp32c3"
      reason: "None"
  depends_components:
    - if: IDF_VERSION == "1.2.3" or IDF_VERSION_MAJOR == 4
      content: [ "VVV" ]
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]
    - default: ["some_1", "some_2", "some_3"]

""",
        encoding='utf8',
    )

    manifest = Manifest.from_file(yaml_file)
    manifest_rule = manifest.rules[0]
    assert (
        repr(manifest_rule.enable) == """[IfClause(stmt='IDF_TARGET == "esp32c3"', temporary=False, reason='None')]"""
    )
    assert (
        repr(manifest_rule.depends_components)
        == """SwitchClause(if_clauses=[IfClause(stmt='IDF_VERSION == "1.2.3" or IDF_VERSION_MAJOR == 4', temporary=False, reason=None), IfClause(stmt='CONFIG_NAME == "AAA"', temporary=False, reason=None)], contents=[['VVV'], ['AAA']], default_clause=['some_1', 'some_2', 'some_3'])"""  # noqa
    )


def test_manifest_switch_clause(tmp_path):
    yaml_file = tmp_path / 'test.yml'
    from idf_build_apps.constants import (
        IDF_VERSION,
    )

    yaml_file.write_text(
        f"""
test1:
  depends_components:
    - if: IDF_VERSION == "{IDF_VERSION}"
      content: [ "VVV" ]
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]
    - default: ["some_1", "some_2", "some_3"]

test2:
  depends_components:
    - if: IDF_TARGET == "esp32"
      content: [ "esp32" ]
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]
    - if: IDF_VERSION == "{IDF_VERSION}"
      content: [ "VVV" ]
    - default: ["some_1", "some_2", "some_3"]

test3:
  depends_components:
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]
    - if: CONFIG_NAME == "BBB"
      content: [ "BBB" ]
    - default: ["some_1", "some_2", "some_3"]

test4:
  depends_components:
    - if: CONFIG_NAME == "BBB"
      content: [ "BBB" ]
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]

test5:
  depends_components:
    - "some_1"
    - "some_2"
    - "some_3"

""",
        encoding='utf8',
    )

    manifest = Manifest.from_file(yaml_file)

    assert manifest.depends_components('test1', None, None) == ['VVV']
    assert manifest.depends_components('test1', None, 'AAA') == ['VVV']

    assert manifest.depends_components('test2', 'esp32', None) == ['esp32']
    assert manifest.depends_components('test2', None, 'AAA') == ['AAA']
    assert manifest.depends_components('test2', 'esp32', 'AAA') == ['esp32']
    assert manifest.depends_components('test2', None, None) == ['VVV']

    assert manifest.depends_components('test3', 'esp32', 'AAA') == ['AAA']
    assert manifest.depends_components('test3', 'esp32', 'BBB') == ['BBB']
    assert manifest.depends_components('test3', 'esp32', '123123') == ['some_1', 'some_2', 'some_3']
    assert manifest.depends_components('test3', None, None) == ['some_1', 'some_2', 'some_3']

    assert manifest.depends_components('test4', 'esp32', 'AAA') == ['AAA']
    assert manifest.depends_components('test4', 'esp32', 'BBB') == ['BBB']
    assert manifest.depends_components('test4', 'esp32', '123123') == []
    assert manifest.depends_components('test4', None, None) == []

    assert manifest.depends_components('test5', 'esp32', 'AAA') == ['some_1', 'some_2', 'some_3']
    assert manifest.depends_components('test5', 'esp32', 'BBB') == ['some_1', 'some_2', 'some_3']
    assert manifest.depends_components('test5', 'esp32', '123123') == ['some_1', 'some_2', 'some_3']
    assert manifest.depends_components('test5', None, None) == ['some_1', 'some_2', 'some_3']


def test_manifest_switch_clause_with_postfix(tmp_path):
    yaml_file = tmp_path / 'test.yml'

    yaml_file.write_text(
        """
.test: &test
  depends_components+:
    - if: CONFIG_NAME == "AAA"
      content: ["NEW_AAA"]
    - if: CONFIG_NAME == "BBB"
      content: ["NEW_BBB"]
  depends_components-:
    - if: CONFIG_NAME == "CCC"

test1:
  <<: *test
  depends_components:
    - if: CONFIG_NAME == "AAA"
      content: [ "AAA" ]
    - if: CONFIG_NAME == "CCC"
      content: [ "CCC" ]
    - default: ["DF"]
""",
        encoding='utf8',
    )
    manifest = Manifest.from_file(yaml_file, root_path=tmp_path)

    assert manifest.depends_components('test1', None, None) == ['DF']
    assert manifest.depends_components('test1', None, 'CCC') == ['DF']
    assert manifest.depends_components('test1', None, 'AAA') == ['NEW_AAA']
    assert manifest.depends_components('test1', None, 'BBB') == ['NEW_BBB']


def test_manifest_switch_clause_wrong_manifest_format(tmp_path):
    yaml_file = tmp_path / 'test.yml'
    from idf_build_apps.constants import (
        IDF_VERSION,
    )

    yaml_file.write_text(
        f"""
    test1:
      depends_components:
        - if: IDF_VERSION == "{IDF_VERSION}"
          content: [ "VVV" ]
        - default: ["some_1", "some_2", "some_3"]
        - hello: 123

    """,
        encoding='utf8',
    )
    try:
        with pytest.warns(UserWarning, match='Folder ".+" does not exist. Please check your manifest file'):
            Manifest.from_file(yaml_file)
    except InvalidManifest as e:
        assert str(e) == "Only the 'if' and 'default' keywords are supported in switch clause."

    yaml_file.write_text(
        f"""
        test1:
          depends_components:
            - if: IDF_VERSION == "{IDF_VERSION}"
              content: [ "VVV" ]
            - default: ["some_1", "some_2", "some_3"]
            - 123
            - 234
        """,
        encoding='utf8',
    )
    try:
        Manifest.from_file(yaml_file)
    except InvalidManifest as e:
        assert str(e) == 'Current manifest format has to fit either the switch format or the list format.'


def test_manifest_with_anchor(tmp_path, monkeypatch):
    yaml_file = tmp_path / 'test.yml'
    yaml_file.write_text(
        """
.base: &base
  depends_components:
    - a

foo: &foo
  <<: *base
  disable:
    - if: IDF_TARGET == "esp32"

bar:
  <<: *foo
""",
        encoding='utf8',
    )

    monkeypatch.setattr(idf_build_apps.manifest.manifest.FolderRule, 'DEFAULT_BUILD_TARGETS', ['esp32'])
    manifest = Manifest.from_file(yaml_file)
    assert manifest.enable_build_targets('bar') == []


def test_manifest_with_anchor_and_postfix(tmp_path):
    yaml_file = tmp_path / 'test.yml'

    yaml_file.write_text(
        """
foo:
""",
        encoding='utf8',
    )
    manifest = Manifest.from_file(yaml_file)
    assert manifest.enable_build_targets('foo') == sorted(SUPPORTED_TARGETS)

    yaml_file.write_text(
        """
.base_depends_components: &base-depends-components
  depends_components:
    - esp_hw_support
    - esp_rom
    - esp_wifi

examples/wifi/coexist:
  <<: *base-depends-components
  depends_components+:
    - esp_coex
  depends_components-:
    - esp_rom
""",
        encoding='utf8',
    )

    manifest = Manifest.from_file(yaml_file)
    assert manifest.depends_components('examples/wifi/coexist') == ['esp_coex', 'esp_hw_support', 'esp_wifi']

    yaml_file.write_text(
        """
.base: &base
  enable:
    - if: IDF_VERSION == "5.2.0"
    - if: IDF_VERSION == "5.3.0"

foo:
  <<: *base
  enable+:
    - if: IDF_VERSION == "5.2.0"
      temp: true
    - if: IDF_VERSION == "5.4.0"
      reason: bar
""",
        encoding='utf8',
    )
    s_manifest_dict = parse(yaml_file)

    yaml_file.write_text(
        """
foo:
    enable:
        -   if: IDF_VERSION == "5.3.0"
        -   if: IDF_VERSION == "5.2.0"
            temp: true
        -   if: IDF_VERSION == "5.4.0"
            reason: bar
""",
        encoding='utf8',
    )
    with open(yaml_file) as f:
        manifest_dict = yaml.safe_load(f) or {}

    assert s_manifest_dict['foo'] == manifest_dict['foo']

    yaml_file.write_text(
        """
.base: &base
  enable:
    - if: IDF_VERSION == "5.3.0"

foo:
  <<: *base
  enable+:
    - if: IDF_VERSION == "5.2.0"
      temp: true
    - if: IDF_VERSION == "5.4.0"
      reason: bar
""",
        encoding='utf8',
    )

    s_manifest_dict = parse(yaml_file)

    yaml_file.write_text(
        """
foo:
    enable:
        -   if: IDF_VERSION == "5.3.0"
        -   if: IDF_VERSION == "5.2.0"
            temp: true
        -   if: IDF_VERSION == "5.4.0"
            reason: bar
""",
        encoding='utf8',
    )
    with open(yaml_file) as f:
        manifest_dict = yaml.safe_load(f) or {}

    assert s_manifest_dict['foo'] == manifest_dict['foo']

    yaml_file.write_text(
        """
.test: &test
  depends_components:
    - a
    - b

foo:
  <<: *test
  depends_components+:
    - c

bar:
  <<: *test
  depends_components+:
    - d
""",
        encoding='utf8',
    )
    s_manifest_dict = parse(yaml_file)
    foo = s_manifest_dict['foo']
    bar = s_manifest_dict['bar']
    assert foo['depends_components'] == ['a', 'b', 'c']
    assert bar['depends_components'] == ['a', 'b', 'd']
    assert id(foo['depends_components']) != id(bar['depends_components'])

    yaml_file.write_text(
        """
.test: &test
  depends_components:
    - if: 1
      value: 123

foo:
  <<: *test
  depends_components+:
    - if: 2
      value: 234

bar:
  <<: *test
  depends_components+:
    - if: 2
      value: 345
""",
        encoding='utf8',
    )
    s_manifest_dict = parse(yaml_file)
    foo = s_manifest_dict['foo']
    bar = s_manifest_dict['bar']
    assert id(foo['depends_components']) != id(bar['depends_components'])
    print(s_manifest_dict)


def test_manifest_postfix_order(tmp_path):
    yaml_file = tmp_path / 'test.yml'
    yaml_file.write_text(
        """
.base_depends_components: &base-depends-components
  depends_components:
    - esp_hw_support

examples/wifi/coexist:
  <<: *base-depends-components
  depends_components+:
    - esp_coex
  depends_components-:
    - esp_coex
""",
        encoding='utf8',
    )

    manifest = Manifest.from_file(yaml_file)
    assert manifest.depends_components('examples/wifi/coexist') == ['esp_hw_support']


def test_from_files_duplicates(tmp_path, monkeypatch):
    yaml_file_1 = tmp_path / 'test1.yml'
    yaml_file_1.write_text(
        """
foo:
  enable:
    - if: IDF_TARGET == "esp32"
""",
        encoding='utf8',
    )

    yaml_file_2 = tmp_path / 'test2.yml'
    yaml_file_2.write_text(
        """
foo:
    enable:
        - if: IDF_TARGET == "esp32"
""",
        encoding='utf8',
    )

    monkeypatch.setattr(idf_build_apps.manifest.manifest.Manifest, 'CHECK_MANIFEST_RULES', True)
    folder_path = os.path.join(os.getcwd(), 'foo')
    os.makedirs(folder_path)

    with pytest.raises(InvalidManifest, match=f'Folder "{folder_path}" is already defined in {yaml_file_1!s}'):
        Manifest.from_files([str(yaml_file_1), str(yaml_file_2)])

    monkeypatch.setattr(idf_build_apps.manifest.manifest.Manifest, 'CHECK_MANIFEST_RULES', False)
    Manifest.from_files([str(yaml_file_1), str(yaml_file_2)])


def test_manifest_dump_sha(tmp_path, sha_of_enable_only_esp32):
    yaml_file = tmp_path / 'test.yml'
    yaml_file.write_text(
        """
foo:
  enable:
    - if: IDF_TARGET == "esp32"
bar:
  enable:
    - if: IDF_TARGET == "esp32"
""",
        encoding='utf8',
    )

    Manifest.from_file(yaml_file).dump_sha_values(str(tmp_path / '.sha'))

    with open('.sha') as f:
        assert f.readline() == f'bar:{sha_of_enable_only_esp32}\n'
        assert f.readline() == f'foo:{sha_of_enable_only_esp32}\n'


def test_manifest_diff_sha(tmp_path, sha_of_enable_only_esp32):
    yaml_file = tmp_path / 'test.yml'
    yaml_file.write_text(
        """
foo:
  enable:
    - if: IDF_TARGET == "esp32"
    - if: IDF_TARGET == "esp32c3"
bar:
  enable:
    - if: IDF_TARGET == "esp32"
baz:
  enable:
    - if: IDF_TARGET == "esp32"
""",
        encoding='utf8',
    )

    with open('.sha', 'w') as fw:
        fw.write(f'bar:{sha_of_enable_only_esp32}\n')
        fw.write('\n')  # test empty line
        fw.write('       ')  # test spaces
        fw.write(f'foo:{sha_of_enable_only_esp32}\n')

    assert Manifest.from_file(yaml_file).diff_sha_with_filepath(str(tmp_path / '.sha')) == {
        'baz',
        'foo',
    }


class TestIfParser:
    def test_invalid_if_statement(self):
        statement = '1'
        with pytest.raises(InvalidIfClause, match='Invalid if clause: 1'):
            IfClause(statement)

    def test_temporary_must_with_reason(self):
        with pytest.raises(InvalidIfClause, match='"reason" must be set when "temporary: true"'):
            IfClause(stmt='IDF_TARGET == "esp32"', temporary=True)



# SPDX-FileCopyrightText: 2023-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import contextlib
import os

import pytest
from pydantic import (
    ValidationError,
)

from idf_build_apps import (
    AppDeserializer,
    CMakeApp,
    MakeApp,
)
from idf_build_apps.main import (
    json_to_app,
)
from idf_build_apps.utils import Literal


def test_serialization():
    a = CMakeApp('foo', 'bar')
    a_s = a.to_json()

    b = CMakeApp.model_validate_json(a_s)
    assert a == b


def test_deserialization():
    a = CMakeApp('foo', 'bar', size_json_filename='size.json')
    b = MakeApp('foo', 'bar', build_log_filename='build.log')

    assert a != b

    with open('test.txt', 'w') as fw:
        fw.write(a.to_json() + '\n')
        fw.write(b.to_json() + '\n')

    with open('test.txt') as fr:
        a_s = AppDeserializer.from_json(fr.readline())
        b_s = AppDeserializer.from_json(fr.readline())

    assert a == a_s
    assert b == b_s


def test_app_sorting():
    a = CMakeApp('foo', 'esp32')
    b = MakeApp('foo', 'esp32')

    c = CMakeApp('foo', 'esp32', size_json_filename='size.json')
    d = CMakeApp('foo', 'esp32s2')
    e = CMakeApp('foo', 'esp32s2', build_comment='build_comment')

    with pytest.raises(TypeError, match="'<' not supported between instances of 'CMakeApp' and 'MakeApp'"):
        assert a < b

    assert a < c < d
    assert d > c > a

    # __EQ_IGNORE_FIELDS__
    assert d == e
    assert not d < e
    assert not d > e


def test_app_deserializer():
    a = CMakeApp('foo', 'esp32')
    b = MakeApp('foo', 'esp32')

    class CustomApp(CMakeApp):
        build_system: Literal['custom'] = 'custom'  # type: ignore

    c = CustomApp('foo', 'esp32')

    assert json_to_app(a.to_json()) == a
    assert json_to_app(b.to_json()) == b

    with pytest.raises(
        ValidationError,
        match="Input tag 'custom' found using 'build_system' does not match "
        "any of the expected tags: 'unknown', 'cmake', 'make'",
    ):
        assert json_to_app(c.to_json()) == c

    assert json_to_app(c.to_json(), extra_classes=[CustomApp]) == c


def test_app_init_from_another():
    a = CMakeApp('foo', 'esp32', build_dir='build_@t_')
    b = CMakeApp.from_another(a, target='esp32c3')

    assert isinstance(b, CMakeApp)
    assert a.target == 'esp32'
    assert b.target == 'esp32c3'
    assert 'build_esp32_' == a.build_dir
    assert 'build_esp32c3_' == b.build_dir


def test_app_hash():
    a = CMakeApp('foo', 'esp32')
    b = CMakeApp('foo/', 'esp32')
    assert a == b
    assert hash(a) == hash(b)
    assert len(list({a, b})) == 1

    with contextlib.chdir(os.path.expanduser('~')):
        a = CMakeApp('foo', 'esp32')
        b = CMakeApp(os.path.join('~', 'foo'), 'esp32')
        assert a == b
        assert hash(a) == hash(b)
        assert len(list({a, b})) == 1



# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
import sys
from pathlib import Path

import pytest

from idf_build_apps.main import main
from idf_build_apps.utils import InvalidCommand


@pytest.mark.parametrize(
    'args, expected_error',
    [
        ([], 'the following arguments are required: --manifest-files, --output/-o'),
        (['--output', 'test.sha1'], 'the following arguments are required: --manifest-files'),
        (['--manifest-files', 'test.yml'], 'the following arguments are required: --output'),
        (['--manifest-files', 'test.yml', '--output', 'test.sha1'], None),
    ],
)
def test_manifest_dump_sha_values(
    args, expected_error, sha_of_enable_only_esp32, sha_of_enable_esp32_or_esp32s2, capsys, monkeypatch
):
    Path('test.yml').write_text(
        """
foo:
  enable:
    - if: IDF_TARGET == "esp32"
bar:
  enable:
    - if: IDF_TARGET == "esp32"
baz:
  enable:
    - if: IDF_TARGET == "esp32"
foobar:
  enable:
    - if: IDF_TARGET == "esp32" or IDF_TARGET == "esp32s2"
""",
        encoding='utf8',
    )

    try:
        with monkeypatch.context() as m:
            m.setattr(sys, 'argv', ['idf-build-apps', 'dump-manifest-sha', *args])

            if not expected_error:
                main()
            else:
                main()
    except SystemExit as e:
        if isinstance(e, InvalidCommand):
            actual_error = e.args[0]
        else:
            actual_error = capsys.readouterr().err
    else:
        actual_error = None

    if expected_error:
        assert actual_error is not None
        assert expected_error in str(actual_error)
    else:
        with open('test.sha1') as fr:
            assert fr.read() == (
                f'bar:{sha_of_enable_only_esp32}\n'
                f'baz:{sha_of_enable_only_esp32}\n'
                f'foo:{sha_of_enable_only_esp32}\n'
                f'foobar:{sha_of_enable_esp32_or_esp32s2}\n'
            )


def test_manifest_patterns(tmp_path, monkeypatch, capsys):
    manifest = tmp_path / 'manifest.yml'
    manifest.write_text(
        """foo:
  enable:
    - if: IDF_TARGET == "esp32"
bar:
  enable:
    - if: IDF_TARGET == "esp32"
        """
    )

    with pytest.raises(SystemExit) as e:
        with monkeypatch.context() as m:
            m.setattr(
                sys,
                'argv',
                ['idf-build-apps', 'find', '--manifest-filepatterns', '**.whatever', '**/manifest.yml', '-vv'],
            )
            main()
        assert e.retcode == 0

    _, err = capsys.readouterr()
    assert f'Loading manifest file {os.path.join(tmp_path, "manifest.yml")}' in err
    assert f'"{os.path.join(tmp_path, "foo")}" does not exist' in err
    assert f'"{os.path.join(tmp_path, "bar")}" does not exist' in err



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

from .report import (
    TestCase,
    TestReport,
    TestSuite,
)

__all__ = ['TestCase', 'TestReport', 'TestSuite']



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
Manifest file
"""

from esp_bool_parser import register_addition_attribute

from .manifest import FolderRule


def folder_rule_attr(target, **kwargs):
    return 1 if target in FolderRule.DEFAULT_BUILD_TARGETS else 0


register_addition_attribute('INCLUDE_DEFAULT', folder_rule_attr)



# SPDX-FileCopyrightText: 2023-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import typing as t
from datetime import datetime

from rich import get_console
from rich._log_render import LogRender
from rich.console import Console, ConsoleRenderable
from rich.containers import Renderables
from rich.logging import RichHandler
from rich.text import Text, TextType


class _OneLineLogRender(LogRender):
    def __call__(  # type: ignore # the original method returns Table instead of Text
        self,
        console: Console,
        renderables: t.Iterable[ConsoleRenderable],
        log_time: t.Optional[datetime] = None,
        time_format: t.Optional[t.Union[str, t.Callable[[datetime], Text]]] = None,
        level: TextType = '',
        path: t.Optional[str] = None,
        line_no: t.Optional[int] = None,
        link_path: t.Optional[str] = None,
    ) -> Text:
        output = Text(no_wrap=True)
        if self.show_time:
            log_time = log_time or console.get_datetime()
            time_format = time_format or self.time_format
            if callable(time_format):
                log_time_display = time_format(log_time)
            else:
                log_time_display = Text(log_time.strftime(time_format), style='log.time')
            if log_time_display == self._last_time and self.omit_repeated_times:
                output.append(' ' * len(log_time_display), style='log.time')
            else:
                output.append(log_time_display)
                self._last_time = log_time_display
            output.pad_right(1)

        if self.show_level:
            output.append(level)
            if self.level_width:
                output.pad_right(max(1, self.level_width - len(level)))
            else:
                output.pad_right(1)

        for renderable in Renderables(renderables):  # type: ignore
            if isinstance(renderable, Text):
                renderable.stylize('log.message')

            output.append(renderable)
            output.pad_right(1)

        if self.show_path and path:
            path_text = Text(style='log.path')
            path_text.append(path, style=f'link file://{link_path}' if link_path else '')
            if line_no:
                path_text.append(':')
                path_text.append(
                    f'{line_no}',
                    style=f'link file://{link_path}#{line_no}' if link_path else '',
                )
            output.append(path_text)
            output.pad_right(1)

        output.rstrip()
        return output


def get_rich_log_handler(level: int = logging.WARNING, no_color: bool = False) -> RichHandler:
    console = get_console()
    console.soft_wrap = True
    console.no_color = no_color
    console.stderr = True

    handler = RichHandler(
        level,
        console,
    )
    handler._log_render = _OneLineLogRender(
        show_level=True,
        show_path=False,
        omit_repeated_times=False,
    )

    return handler


def setup_logging(verbose: int = 0, log_file: t.Optional[str] = None, colored: bool = True) -> None:
    """
    Setup logging stream handler

    :param verbose: 0 - WARNING, 1 - INFO, 2 - DEBUG
    :param log_file: log file path
    :param colored: colored output or not
    :return: None
    """
    if not verbose:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG

    package_logger = logging.getLogger(__package__)
    package_logger.setLevel(level)

    if log_file:
        handler: logging.Handler = logging.FileHandler(log_file)
    else:
        handler = get_rich_log_handler(level, not colored)
    if package_logger.hasHandlers():
        package_logger.handlers.clear()
    package_logger.addHandler(handler)

    package_logger.propagate = False  # don't propagate to root logger



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import re
import typing as t

LOGGER = logging.getLogger(__name__)


class SessionArgs:
    workdir: str = os.getcwd()
    override_sdkconfig_items: t.Dict[str, t.Any] = {}
    override_sdkconfig_file_path: t.Optional[str] = None

    def set(self, parsed_args, *, workdir=None):
        if workdir:
            self.workdir = workdir
        self._setup_override_sdkconfig(parsed_args)

    def clean(self):
        self.override_sdkconfig_items = {}
        self.override_sdkconfig_file_path = None

    def _setup_override_sdkconfig(self, args):
        override_sdkconfig_items = self._get_override_sdkconfig_items(
            args.override_sdkconfig_items.split(',') if args.override_sdkconfig_items else ()
        )
        override_sdkconfig_files_items = self._get_override_sdkconfig_files_items(
            args.override_sdkconfig_files.split(',') if args.override_sdkconfig_files else ()
        )

        override_sdkconfig_files_items.update(override_sdkconfig_items)
        self.override_sdkconfig_items = override_sdkconfig_files_items

        override_sdkconfig_merged_file = self._create_override_sdkconfig_merged_file(self.override_sdkconfig_items)
        self.override_sdkconfig_file_path = override_sdkconfig_merged_file

    def _get_override_sdkconfig_files_items(self, override_sdkconfig_files: t.Tuple[str]) -> t.Dict:
        d = {}
        for f in override_sdkconfig_files:
            # use filepath if abs/rel already point to itself
            if not os.path.isfile(f):
                # find it in the workdir
                LOGGER.debug('override sdkconfig file %s not found, checking under app_dir...', f)
                f = os.path.join(self.workdir, f)
                if not os.path.isfile(f):
                    LOGGER.debug('override sdkconfig file %s not found, skipping...', f)
                    continue

            with open(f) as fr:
                for line in fr:
                    m = re.compile(r'^([^=]+)=([^\n]*)\n*$').match(line)
                    if not m:
                        continue
                    d[m.group(1)] = m.group(2)
        return d

    def _get_override_sdkconfig_items(self, override_sdkconfig_items: t.Tuple[str]) -> t.Dict:
        d = {}
        for line in override_sdkconfig_items:
            m = re.compile(r'^([^=]+)=([^\n]*)\n*$').match(line)
            if m:
                d[m.group(1)] = m.group(2)
        return d

    def _create_override_sdkconfig_merged_file(self, override_sdkconfig_merged_items) -> t.Optional[str]:
        if not override_sdkconfig_merged_items:
            return None
        f_path = os.path.join(self.workdir, 'override-result.sdkconfig')
        with open(f_path, 'w+') as f:
            for key, value in override_sdkconfig_merged_items.items():
                f.write(f'{key}={value}\n')
        return f_path
